"""Mushroom item class facility."""

from ..logconfig import APP_LOGGER
from ..appcommon.anim.player.main import AnimationPlayer
from ..appcommon.collision import does_touch_screen
from ..frontprop.toxicpollencloud import ToxicPollenCloud


logger = APP_LOGGER.getChild(__name__)


class Mushroom:
    """A mushroom equipabble item.

    It's a projectile item whose orientation (left or right)
    is controlled by the player orientation.

    It is thrown a little over the player head and its
    trajectory is controlled by the gravity, making it
    form arc like trajectory.

    Whenever it hits a block or actor it bursts into a
    circular-shaped cloud which deals a little damage
    in touching actors.

    It uses an update_manager analogy to manage different
    scenarios: mushroom on plant vs mushroom being used.
    When on inventory, the object doesn't have its state
    updated on each loop, just stays there not being updated
    until it is finally used, switching to the proper state.
    """

    def __init__(self, prop_name, coordinates, level, plant=None):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates
            representing the desired topleft coordinates
            for the rect attribute.
        level
            Instance of gamelevel.main.Level class.
        plant
            If the mushroom was instantiated on a mushroom
            plant, then a reference to the plant instance
            must be provided.
        """
        self.prop_name = prop_name

        self.anim_player = AnimationPlayer(
            self,
            prop_name,
            coordinates_name="bottomleft",
            coordinates_value=coordinates,
        )

        self.level = level
        self.plant = plant
        self.set_operation_structure()

    def set_operation_structure(self):
        """Set other important operational structures.

        Those include atributes, methods and other objects."""
        self.top_speed = 10
        self.x_speed = 0
        self.x_acceleration = 1

        self.y_speed = 0
        self.gravity_accel = 2

        self.facing_right = True
        self.is_burst = False

        self.toxic_pollen_cloud = ToxicPollenCloud(
            "toxic_pollen_cloud", self.rect.topleft, self.level
        )
        self.anim_player.play()

    def prepare_for_plant(self):
        """Make it mushroom plant-ready."""
        self.rect.x += 6
        self.rect.y += 56

        self.anim_player.switch_animation("growing")
        self.update_method = self.update_on_plant

    def prepare_for_usage(self, player):
        """Make it usage ready.

        player
            An instance of player.main.Player class.
        """
        if player.facing_right:
            self.facing_right = True
            self.x_speed = 10
            self.anim_player.switch_animation("thrown")
            self.rect.bottomleft = player.rect.topright

        else:
            self.facing_right = False
            self.x_speed = -10
            self.anim_player.switch_animation("thrown_reverse")
            self.rect.bottomright = player.rect.topleft

        self.y_speed = -26
        self.update_method = self.update_for_usage

    def update(self):
        """Execute method stored in update_method."""
        self.update_method()
        self.manage_animations()
        self.anim_player.update()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()

    ### XXX
    ### entire class needs refactoring, but specially these
    ### two methods below and their dependencies

    def update_on_plant(self):
        """Do nothing."""

    def update_for_usage(self):
        """Update item when being used by player."""
        self.move_x()
        self.move_y()

    def manage_animations(self):
        """Select specific animation based on state."""
        # XXX
        # Maybe turn this kind of animation branching
        # into a mini language that can be used to set
        # custom behaviours by giving sequences of
        # animations to animation player;
        # push the matter further, it could even be
        # implemented on the .gameobj file too, maybe.

        anim_name = self.anim_player.anim_name

        if anim_name == "growing":

            loops_no = self.anim_player.peek_next_loops_no()

            if loops_no >= 1:
                self.anim_player.switch_animation("idle")
                self.plant.choose_interaction_method()

        elif anim_name == "bursting":

            loops_no = self.anim_player.peek_after_next_loops_no()

            if loops_no >= 1:
                self.spawn_pollen_cloud()
                self.level.remove_obj_from_group(self, "equippable_items", True)

    def move_x(self):
        """Calculate and perform movement in x axis."""
        if self.is_burst:
            return

        self.rect.x += self.x_speed

        for group_name in ("blocks", "actors"):
            collided_prop = self.level.onscreen_map[group_name].collide_any(self)
            if collided_prop:
                self.burst()
                break

        if not self.is_burst:
            if self.rect.colliderect(self.level.player.rect):
                self.burst()

        if not self.is_burst:
            if not does_touch_screen(self):
                self.burst()

    def move_y(self):
        """Calculate and perform movement in y axis."""
        if self.is_burst:
            return

        self.apply_gravity_pull()
        self.rect.y += self.y_speed

        for group_name in ("blocks", "actors"):
            collided_prop = self.level.onscreen_map[group_name].collide_any(self)
            if collided_prop:
                self.burst()
                break

        if not self.is_burst:
            if self.rect.colliderect(self.level.player.rect):
                self.burst()

        if not self.is_burst:
            if not does_touch_screen(self):
                self.burst()

    def apply_gravity_pull(self):
        """Move player according to gravity force."""
        will_traverse_floor = self.y_speed > self.rect.height

        if not will_traverse_floor:
            self.y_speed += self.gravity_accel

    def grow(self):
        """Set the current state to 'growing'."""
        self.anim_player.switch_animation("growing")

    def burst(self):
        """Burst the item."""
        self.is_burst = True
        self.anim_player.switch_animation("bursting")

    def spawn_pollen_cloud(self):
        """Insert toxic pollen cloud into level."""
        self.toxic_pollen_cloud.prepare_to_enter(self)

        self.level.add_obj_to_group(self.toxic_pollen_cloud, "front_props")

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for the back_props layer."""

### local imports

from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.surf import render_image


class VinesLadder(BasicObject):
    """A ladder made of vines.

    Objects whose prop_name attibute contain 'ladder'
    string are treated by player.main.Player as climbable.
    """

    level = None

    prop_map = {
        "vines_climbable": render_image("vines_climbable.png"),
        "vines_top": render_image("vines_top.png"),
        "vines_grid_climbable": render_image("vines_grid_climbable.png"),
        "vines_grid_top": render_image("vines_grid_top.png"),
    }

    def __init__(self, prop_name, coordinates):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        """
        self.prop_name = prop_name

        prop_value = VinesLadder.prop_map[prop_name]

        # TODO this load_image wasn't even imported; this
        # must be code left by mistake here; review it
        # (along with entire module); in fact, there should
        # be a package-wide refactoring soon;
        if isinstance(prop_value, str):
            prop_value = load_image(prop_value).convert_alpha()
            VinesLadder.prop_map[prop_name] = prop_value

        self.image = prop_value

        self.rect = self.image.get_rect()

        self.rect.bottomleft = coordinates

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)


class SimpleDoor(BasicObject):
    """A simple door. Portal like behaviour."""

    level = None

    prop_map = {"simple_door": render_image("simple_door.png")}

    def __init__(self, prop_name, coordinates, next_level, portal_name, destination):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        next_level
            A string with the basename of the level filepath.
        portal_name
            A string representing the specific portal name of
            this door instance (entrance). Used to identify
            the portal.
        destination
            A string representing the specific portal name of
            the portal on which we will end up exiting. Used
            to identify such portal.
        """
        self.prop_name = prop_name

        self.image = SimpleDoor.prop_map[prop_name]

        self.rect = self.image.get_rect()
        self.rect.bottomleft = coordinates

        self.next_level = next_level
        self.portal_name = portal_name
        self.destination = destination

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)


class CaveEntrance(BasicObject):
    """A cave entrance. Portal like behaviour."""

    level = None

    prop_map = {"cave_entrance": render_image("cave_entrance.png")}

    def __init__(self, prop_name, coordinates, next_level, portal_name, destination):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        next_level
            A string with the basename of the level filepath.
        portal_name
            A string representing the specific portal name of
            this door instance (entrance). Used to identify
            the portal.
        destination
            A string representing the specific portal name of
            the portal on which we will end up exiting. Used
            to identify such portal.
        """

        self.prop_name = prop_name

        self.image = CaveEntrance.prop_map[prop_name]

        self.rect = self.image.get_rect()
        self.rect.bottomleft = coordinates

        self.next_level = next_level
        self.portal_name = portal_name
        self.destination = destination

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for new game creation.

Here we provide an object which manages the screen.
It's purpose is to present the player with means to
enter a name for a new save to be created.

Once the name is entered and approved, the player starts
to play the first level."""

from os import listdir
from random import seed, choice
from string import ascii_lowercase, digits
from itertools import cycle

### third-party imports

from pygame import QUIT, KEYDOWN, K_ESCAPE, K_BACKSPACE, K_RETURN

from pygame.display import update
from pygame.event import get as get_events

### local imports

from ..config import SCREEN, SCREEN_RECT, SAVES_DIR, GAME_REFS

from ..appcommon.task import update_task_manager
from ..appcommon.surf import INVISIBLE_SURF, render_rect
from ..appcommon.dialog import cache_message, display_message
from ..appcommon.autoblit import BlitterSet
from ..appcommon.text.main import render_text
from ..appcommon.exception import (
    QuitGameException,
    LevelSwitchException,
    ManagerSwitchException,
)

from ..appcommon.behaviour.save import setup_new_save
from ..appcommon.behaviour.load import load_game

from ..sound import SOUNDS_MAP
from ..palette import WHITE, SKY_COLOR


CHARACTERS = ascii_lowercase + digits


class NewGameWidget(object):
    """Widget for setting up a new game and its data.

    The user is prompted for a name for a new save."""

    def __init__(self, main_menu):
        """Set variables and perform setups.

        main_menu
            The instance of the mainmenu.main.MainMenu class.
        """
        self.main_menu = main_menu

        self.error_sound = SOUNDS_MAP["error"]

        self.set_labels()
        self.background = render_rect(*SCREEN_RECT.size, color=SKY_COLOR)

        self.prompt = TextPrompt((520, 340), self)

        cache_message("three_or_more", "Save name must have 03 or more characters.")
        cache_message("name_exists", "The name you picked already exists.")
        cache_message("ten_or_less", "Save name must have 10 or less characters.")

    def set_labels(self):
        """Create a sprite to serve as a label."""

        label_texts = [
            "Please, enter name for new save slot:",
            "use lowercase letters and numbers;",
            "the name must have from 03 to 10 characters;",
            "press <ENTER> when done;",
            "press <BACKSPACE> to delete last character;",
        ]

        self.labels = BlitterSet()
        y = 40
        for text in label_texts:

            label = render_text(
                text,
                font_size=32,
                foreground_color=WHITE,
                return_obj=True,
                coordinates_value=(60, y),
            )

            self.labels.add(label)
            y += 40

    def control(self):
        """Update widget state based on event_queue."""
        for event in get_events():
            if event.type == QUIT:
                raise QuitGameException

            ### Keyboard
            elif event.type == KEYDOWN:

                if event.unicode in CHARACTERS:
                    self.prompt.insert_char(event.unicode)

                elif event.key == K_BACKSPACE:
                    self.prompt.delete_last_char()

                elif event.key == K_RETURN:
                    self.create_new_game()

                elif event.key == K_ESCAPE:
                    raise ManagerSwitchException(self.main_menu)

    def update(self):
        """Update the task manager."""
        update_task_manager()

    def draw(self):
        """Draw widget objects."""
        SCREEN.blit(self.background, (0, 0))
        self.labels.draw()
        self.prompt.draw()

        update()

    def create_new_game(self):
        """Retrieve string from prompt to create new save."""
        save_name = self.prompt.save_name

        if len(save_name) <= 2:

            self.error_sound.play()
            display_message("three_or_more", self)

        else:

            dirname = "save_" + save_name

            if dirname in listdir(SAVES_DIR):

                self.error_sound.play()
                display_message("name_exists", self)

            else:

                setup_new_save(dirname)

                SOUNDS_MAP["game_start"].play()

                GAME_REFS.dirname = dirname
                level_name = load_game()

                raise LevelSwitchException(level_name)


class TextPrompt:
    """A text prompt which displays input."""

    def __init__(self, topleft, new_game_widget):
        """Initialize superclass and set instance.

        topleft
            the topleft coordinates to begin blit
            surfaces in the update method.
        new_game_widget
            reference to mainmenu.newgame.NewGameWidget class.
        """
        self.topleft = topleft
        self.new_game_widget = new_game_widget

        self.save_name = ""

        seed()
        self.typing_sounds = [SOUNDS_MAP["key_typing01"], SOUNDS_MAP["key_typing02"]]

        self.build_surface_map()
        self.blinking_cursor = BlinkingCursor()

    def build_surface_map(self):
        """Create and store text surfaces each character."""
        self.surface_map = {}
        for char in CHARACTERS:
            text_surf = render_text(char, font_size=56, foreground_color=WHITE)
            self.surface_map[char] = text_surf

    def draw(self):
        """Update widget state."""
        x, y = self.topleft

        for char in self.save_name:
            surf = self.surface_map[char]
            width = surf.get_width()
            SCREEN.blit(self.surface_map[char], (x, y))
            x += width

        self.blinking_cursor.blit_on_topleft((x, y))

    def delete_last_char(self):
        """Delete last char in save name."""
        if self.save_name:
            self.save_name = self.save_name[:-1]

    def insert_char(self, char):
        """Replace first '_' character by provided char.

        char
            A string of length == 1.
        """
        if len(self.save_name) < 10:
            self.save_name = self.save_name + char
            choice(self.typing_sounds).play()

        else:
            SOUNDS_MAP["error"].play()
            display_message("ten_or_less", self.new_game_widget)


class BlinkingCursor:
    """A blinking cursor to assist the text prompt."""

    def __init__(self):
        """Assign variables and perform setups."""
        self.image = render_rect(40, 70, WHITE)

        ### create callable to return True in the first
        ### n calls, then False in the next n calls, and
        ### so on

        n = 15
        self.blit = cycle((True,) * n + (False,) * n).__next__

    def blit_on_topleft(self, topleft):
        """Blit surface on screen in topleft coordinates.

        topleft
            Topleft coordinates wherein to blit the surface.
        """
        if self.blit():
            SCREEN.blit(self.image, topleft)



"""Player class facility."""

from ..config import SCREEN, SCREEN_WIDTH, SCREEN_HEIGHT
from ..common.behaviour import empty_function
from ..appcommon.surf import render_image, INVISIBLE_SURF

# XXX
# art: blit a backpack besides the items
# and also, maybe, blit a semi-transparent
# black background behind the items to
# give a bit contrast with the game world

SURFACE_MAP = {"mushroom": render_image("mushroom_topleft.png")}


class EquippedDisplay:
    """A display for equipped items."""

    def __init__(self):
        """Initialize superclass and set variables."""
        self.image = INVISIBLE_SURF
        self.rect = self.image.get_rect()

        self.perform_setup()

        self.update = empty_function

    def perform_setup(self):
        """Set useful variables for operation."""
        self.current_item_name = None
        self.current_item_count = 0
        self.positions = []

    def draw(self):
        """Update object state."""
        if self.current_item_count:
            for i in range(self.current_item_count):
                SCREEN.blit(self.image, self.positions[i])

    def update_visual_data(self, item_name, count):
        """Update surfaces whenever item or count changes.

        item_name
            A string representing a item name.
        count
            An integer. Represents how many items
            are on the inventory.
        """
        if self.current_item_name != item_name:
            self.current_item_name = item_name
            self.image = SURFACE_MAP[item_name]
            self.calculate_surface_positions()
        self.current_item_count = count

    def calculate_surface_positions(self):
        """Calculate positions to blit self.image.

        This takes into account the dimensions of the
        current item surface to give a pleasing padding/
        offset, so the surfaces aren't displayed too far
        or near the edges of the screen since items can have
        different dimensions."""
        self.positions = []

        surf_width, surf_height = self.image.get_size()

        vertical_padding = 32
        y = SCREEN_HEIGHT - (vertical_padding + surf_height)

        horizontal_padding = 32
        x = SCREEN_WIDTH - (horizontal_padding + surf_width)

        surfaces_distance = horizontal_padding + surf_width
        for i in range(6):
            offset = surfaces_distance * i
            position = (x - offset, y)

            self.positions.append(position)



"""Facility for the mushroom plant middle prop class."""

from types import SimpleNamespace

from ..config import SCREEN
from ..common.math import unscroll_coordinates
from ..appcommon.surf import render_image
from ..appcommon.autoblit import BasicObject
from ..appcommon.promptfactory import prompt_factory
from ..appcommon.dialoguefactory import dialoguebox_factory
from ..appcommon.task import add_task, add_sendoff_task
from ..throwing.mushroom import Mushroom


class MushroomPlant(BasicObject):
    """A plant parasited by mushrooms."""

    level = None

    prop_map = {"mushroom_plant": render_image("mushroom_plant.png")}

    def __init__(self, prop_name, coordinates, has_mushroom):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        has_mushroom
            Boolean to flag if the plant has 'spawned' a
            throwing.mushroom.Mushroom
            object.
        """
        self.prop_name = prop_name

        self.image = MushroomPlant.prop_map[prop_name]

        self.rect = self.image.get_rect()
        self.rect.bottomleft = coordinates

        self.has_mushroom = has_mushroom
        self.set_operation_structure()

    def set_operation_structure(self):
        """Set other important operational structures.

        Those include atributes, methods and other objects."""
        ## prepare prompt
        self.object_screen_midtop = unscroll_coordinates(self.rect.midtop)
        # XXX
        # This offset should be made inside the
        # dialogue/prompt mechanism. Ponder. This tied to
        # the notion that dialogue/prompt feature should
        # have references to the objects instead of
        # receiving their position.
        offset_screen_midtop = (
            self.object_screen_midtop[0],
            self.object_screen_midtop[1] - 32,
        )

        # must come after defining self.object_screen_midtop
        self.instantiate_mushroom()

        item_list = []

        # XXX
        # This originally got a reference inside the
        # Mushroom animation_map class attribute.
        # Maybe in the future I might set a feature to
        # return a reference to a specific surface there
        surface = render_image("mushroom_topleft.png")

        item_list.append(surface)

        prompt_pair = prompt_factory(
            "Should I take the item?", offset_screen_midtop, item_list
        )
        prompt_action = self.give_item
        self.prompt = (prompt_pair, prompt_action)

        ## Prepare dialogue
        dialogue = [{"player": "It seems there's no usable mushroom here."}]
        self.dialogue = dialoguebox_factory(dialogue, offset_screen_midtop)

        ## Pick suitable method
        self.choose_interaction_method()

        ## Simple task control
        self.growth_task = SimpleNamespace(finished=True)

    def instantiate_mushroom(self):
        """Instantiate and store a mushroom."""
        self.mushroom = Mushroom(
            "mushroom", self.object_screen_midtop, self.level, self
        )

        self.mushroom.prepare_for_plant()

    def choose_interaction_method(self):
        """Alocate method to use for interactions."""
        if self.has_mushroom:
            self.receive_interaction = self.deliver_prompt

        else:
            self.receive_interaction = self.deliver_dialogue

    def update(self):
        """Update object state."""
        if self.has_mushroom:
            self.mushroom.update()

        else:
            self.set_growth_task()

    def draw(self):
        """Update object state."""
        if self.has_mushroom:
            self.mushroom.draw()

        super().draw()

    def set_growth_task(self):
        """Set growth task if needed."""
        if self.growth_task.finished:
            self.growth_task = add_task(self.set_mushroom_growth, 5000)

    def deliver_prompt(self, player):
        """Send prompt to player.

        player
            The instance of player.main.Player.
        """
        player.send_prompt(self.prompt)

    def deliver_dialogue(self, player):
        """Send dialogue to player.

        player
            The instance of player.main.Player.
        """
        player.send_dialogue(self.dialogue)

    def set_mushroom_growth(self):
        """Make a mushroom grow on plant."""
        self.has_mushroom = True
        self.mushroom.grow()
        self.json_data["has_mushroom"] = True

    def give_item(self):
        """Give item to player and clean variables."""
        try:
            player = self.level.player

        except AttributeError as err:
            # XXX
            # filter this better
            print(err)

        else:
            player.get_item(self.mushroom, self)

    def got_item(self):
        """Make arrangements for itemless state."""
        self.has_mushroom = False

        self.instantiate_mushroom()
        self.json_data["has_mushroom"] = False
        self.choose_interaction_method()

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Game launching."""

### local imports

# XXX maybe wrap the run_game import in a try/exception
# clause, since though rare sometimes refactoring
# errors may occur; ponder

from .logconfig import APP_LOGGER


### instantiating local logger
logger = APP_LOGGER.getChild(__name__)


def main():

    logger.info("Launching game.")
    from .gameloop import run_game

    run_game()


### brief instructions for module execution

if __name__ == "__main__":
    main()



"""Facility for displaying and managing dialogues."""

from ..config import SCREEN
from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.surf import INVISIBLE_SURF


# XXX
# Ideas for improvement:
# The way dialogue manager works is somewhat cumbersome,
# I mean, it's not flexible at all. You can only have
# 2 objects 'speaking' in the same dialogue and their
# positions should be provided in anticipation, what
# prevents the dialogue box to follow moving objects.
# Change this in the future.


class DialogueManager:
    """Display and manages dialogue boxes."""

    def __init__(self, player):
        """Initialize superclass and set variables.

        player
            A reference to the instance of player.main.Player
            class.
        """
        self.player = player
        self.dialogue = None
        self.image = INVISIBLE_SURF.copy()
        self.rect = self.image.get_rect()

    def update(self):
        """Update object state."""
        try:
            self.dialogue.update()

        ### when dialogue attribute is None, due to
        ### trying accessing update attribute
        except AttributeError:
            pass

    def draw(self):
        """Update object state."""
        try:
            self.dialogue.draw()

        ### when dialogue attribute is None, due to
        ### trying accessing draw attribute
        except AttributeError:
            pass

    def get_dialogue(self, dialogue_pairs):
        """Instantiate a dialogue box.

        dialogue_pairs
            A list containing tuples with pairs of
            information. Such information consists
            of position coordinates of whoever is speaking
            and the dialogue surface to be displayed.
            This list is generated in
            appcommon/dialoguefactory.py
        """
        self.dialogue = DialogueDisplay(self, self.player, dialogue_pairs)

    def finish_dialogue(self):
        """Set current dialogue to None."""
        self.dialogue = None

    def skip_dialoguebox(self):
        """Skip current dialoguebox."""
        if self.dialogue:
            self.dialogue.switch_sprite()


class DialogueDisplay(BasicObject):
    """Object switcher for dialogue boxes and monologues."""

    def __init__(self, manager, player, dialoguebox_pairs):
        """Initialize superclass and prepare sprites.

        manager
            The instance of
            player.dialoguemanager.DialogueManager
        player
            A reference to the player.main.Player instance.
        dialoguebox_pairs
            A list containing tuples with pairs of
            information. Such information consists
            of position coordinates of whoever is speaking
            and the dialogue surface to be displayed.
            This list is generated in
            appcommon/dialoguefactory.py
        """
        self.manager = manager
        self.player = player

        self.dialogue_setup(dialoguebox_pairs)

    def dialogue_setup(self, dialoguebox_pairs):
        """Set dialoguebox_pairs to be used.

        dialoguebox_pairs
            A list containing tuples with pairs of
            information. Such information consists
            of position coordinates of whoever is speaking
            and the dialogue surface to be displayed.
            This list is generated in
            appcommon/dialoguefactory.py
        """
        self.dialogue = iter(dialoguebox_pairs)
        self.switch_sprite()

    def update(self):
        """Update object state."""
        if self.is_following_player:
            self.adjust_position()

    def switch_sprite(self):
        """Switch to next sprite, if any, or finish."""
        try:
            dialogue_info = next(self.dialogue)
        except StopIteration:
            self.manager.finish_dialogue()
        else:
            self.image = dialogue_info[1]
            self.rect = self.image.get_rect()

            position = dialogue_info[0]
            if position == "player_pos":
                self.is_following_player = True
                self.adjust_position()
            else:
                self.is_following_player = False
                self.rect.midbottom = position

    def adjust_position(self):
        """Adjust dialoguebox position to follow player."""
        player_midtop = self.player.rect.midtop
        offset_player_midtop = (player_midtop[0], player_midtop[1] - 32)
        self.rect.midbottom = offset_player_midtop



"""Facility for application specific behaviours."""

from os import listdir, remove
from os.path import join

### local imports

from ...logconfig import APP_LOGGER

from ...config import GAME_REFS, SAVES_DIR


### instantiate local logger
logger = APP_LOGGER.getChild(__name__)


def remove_buffers():
    """Remove existing buffers.

    This happens whenever the player quits to main menu
    or game, abandoning the current state/level data,
    present in the buffers. This just means all buffer
    files will be erased (if any).
    """
    logger.debug("Buffers being removed.")

    ### get the current name of the folder representing
    ### the game slot used
    try:
        save_dir = GAME_REFS.dirname

    ### if it is missing, return earlier
    except AttributeError:
        return

    ### Get and remove all buffers (.swp files) in current
    ### save dir

    full_save_dir = join(SAVES_DIR, save_dir)

    buffers = [
        join(SAVES_DIR, save_dir, item)
        for item in listdir(full_save_dir)
        if item.endswith(".swp")
    ]

    for item in buffers:
        remove(item)



"""General utilities for the animation player."""

from abc import ABC
from functools import partial, partialmethod
from operator import attrgetter

### local import
from ....common.behaviour import empty_function, get_nested_value


class Utilities(ABC):
    """Provides utilities for the animation manager."""

    def jump_to_frame(self, frame_index):
        """Jump animation to frame_index by rotating values.

        frame_index (integer)
            Represents the frame to which we want to jump.
            Can be any value, but it will be clamped to the
            animation length range.

        Jumps to an specific frame, regardless of whether
        the animation is being played or paused.
        """
        ### clamp index to allowed range
        # (obs.: if needed elsewhere, turn this section into
        # its own method)

        max_index = self.get_animation_length() - 1
        clamped_index = max(min(max_index, frame_index), 0)

        ### calculate the difference between the desired
        ### frame (the clamped frame index) and the current
        ### frame number; such difference represents
        ### the number of steps to walk
        steps = clamped_index - self.get_current_frame()

        ### finally update the animation values in the nodes
        ### after the specified number of steps
        ### calculated above, that is, until reaching the
        ### desired values; the equivalent of jumping to
        ### the desired frame;
        self.update_nodes(steps)

    def get_animation_length(self):
        """Return length of current animation."""
        return self.anim_clock.length

    def get_current_frame(self):
        """Return current frame of animation clock."""
        return self.anim_clock.get_index_of_first()

    def get_current_loops_no(self):
        """Return number of loops performed."""
        return self.anim_clock.loops_no

    def peek_loops_no(self, steps):
        """Return loops number after temporary rotation.

        steps (integer)
            Number of rotations to go back or forth
            (depending on signal) before peeking.

        We do so by calling an eponymous method in the
        in the animation clock and returning its return
        value.
        """
        return self.anim_clock.peek_loops_no(steps)

    peek_next_loops_no = partialmethod(peek_loops_no, 1)
    peek_after_next_loops_no = partialmethod(peek_loops_no, 2)

    def ensure_animation(self, animation_name):
        """Only switch if given animation != current one.

        Also return the value returned by the
        switch_animation method, which is True when
        the operation succeeds.

        animation_name (string)
            represents the unique name of an animation.
        """
        if animation_name != self.anim_name:
            return self.switch_animation(animation_name)

    ### frame playback options

    def play(self):
        """Start playing animation.

        This is done by setting the step_forth method
        into the update attribute. This is usually the
        desired behaviour, since it will keep looping the
        current animation indefinitely.
        """
        self.update = self.step_forth

    def pause(self):
        """Pause the animation.

        This is done by setting the update_nodes method
        into the update attribute. This means the
        animation will be updated with the same values
        every frame instead of successively walk through
        the animation values, which will make the animation
        appear frozen since we just assign the same values
        again and again, for both surfaces and positions.
        """
        self.update = self.update_nodes

    def is_paused(self):
        """Return True if animation is paused.

        That is, if update attribute is set to
        the update_nodes method.
        """
        return self.update == self.update_nodes

    def toggle_play(self):
        """Toggle behaviour between play/pause state."""
        if self.is_paused():
            self.play()
        else:
            self.pause()

    ### drawing related options

    def check_walking_skipping(self):
        """Check need to prevent walking on first frame.

        When we switch to a new animation when not paused,
        the first execution of the update routine would
        naturally cause the animation values of the first
        frame to be skipped when the timing indices rotate
        using the WalkingDeque.walk method.

        To prevent that, we check if it is the case here
        (the animation is not paused) and if so we set a
        special method, called skip_walking, in lieu of the
        current update routine. Such special method prevents
        the undesired behaviour described above.

        One additional condition, though, is that the
        special method must not be set yet, since setting
        the special method again would cause another
        undesired effect described in the is_skipping
        method.

        Once the conditions above are met, before setting
        the special method, we store a reference to the
        current udpate routine, so it can be restored later.
        """
        ### if not paused nor skipping, store current update
        ### routine and assign the skip_walking in its place
        if not self.is_paused() and not self.is_skipping():

            self.previous_update_routine = self.update

            self.update = self.skip_walking

    def is_skipping(self):
        """Return True if skip_walking is set.

        This is used to prevent trying to set the
        skip_walking method again when it is already set.

        Since the skip_walking method automatically sets the
        update routine to its previous value, if the
        previous value of the update routine was itself,
        the routine would keep setting itself as the update
        routine indefinitely.
        """
        return self.update == self.skip_walking

    def skip_walking(self):
        """Update w/out walking and revert update routine."""
        self.update_nodes()
        self.update = self.previous_update_routine

    def stop(self):
        """Pause and walk to the beginning of animation."""
        self.pause()
        self.go_to_beginning()

    def go_to_beginning(self):
        """Clear rotation and update node values.

        By doing so, it has the same effect as having the
        animation go to the beginning (regardless of whether
        its paused or not).
        """
        self.clear_rotations()
        self.update_nodes()

    def set_draw_routine(self, routine_name):
        """Set drawing routine, return True if succeed.

        routine_name (string)
            Attribute name of the draw routine. It should
            have one of the following values:

            - 'draw_nodes'
                Set the draw_routine to draw only nodes.
                This is the default behaviour.
            - 'draw_nodes_and_bounding_rect'
                Set the draw routine to draw both the
                nodes and also their bounding rects.
            - 'draw_nodes_and_col_rect'
                Set the draw routine to draw both the
                nodes and also their collision rects,
                if they have them.
        """
        ### routine_name value checking

        if routine_name in (
            "draw_nodes",
            "draw_nodes_and_bounding_rect",
            "draw_nodes_and_col_rect",
        ):

            ## routine assignment
            self.draw = getattr(self, routine_name)

            ## indicate success
            return True

    ### data queries

    def get_nodes(self, exclude_root=False, order="node_names", reverse=False):
        """Return nodes for current animation.

        exclude_root (boolean - default is False)
            whether root should be excluded or not in the
            nodes returned.
        order (string - default is 'node_names')
            represents the order in which node references
            must be arranged.
        reverse (boolean - default is False)
            wether to keep the chosen ordering or reverse it.
        """
        ### order value checking

        legal_values = ("node_names", "updating_order", "drawing_order")

        if order in legal_values:

            ### create slice object according to need of
            ### reversing order or not

            last_param = -1 if reverse else None
            slice_obj = slice(None, None, last_param)

            ### retrieve attribute using order and slice
            order = getattr(self, order)[slice_obj]

            ### get the node references
            nodes = map(self.node_map.__getitem__, order)

            ### return nodes with/without root depending on
            ### exclude_root parameter

            if exclude_root:
                return (node for node in nodes if node is not self.root)

            else:
                return nodes

        else:
            raise ValueError("order must be in one of: " + str(legal_values))



"""Facility for time related tools."""

from datetime import datetime


def get_isostring_datetime():
    """Return a iso formated today datetime, including time.

    get_isostring_datetime() -> str

    For instance:
    get_isostring_datetime -> '2018-04-02 14:22:36'

    >>> date_regex = '[0-9]{4}-[01][0-9]-[0-3][0-9]'
    >>> time_regex = '[0-2][0-9]:[0-5][0-9]:[0-5][0-9]'
    >>> full_regex = "{} {}".format(date_regex, time_regex)
    >>> isostr = get_isostring_datetime()
    >>> from re import fullmatch
    >>> bool(fullmatch(full_regex, isostr))
    True
    """
    return datetime.now().strftime("%Y-%m-%d %T")



"""Facility for text related utilities."""

### third-party imports
from pygame import Surface

### local imports

from ..autoblit import BasicObject
from ...palette import BLACK
from ...font import get_font


# XXX maybe create a BasicObject subclass with extra
# text processing capabilities, like auto update its
# image attribute; there's already at least one or two
# cases where this would be useful.


def get_text_size(text, font_size=22, font_style="default", padding=5):
    """Return surf size of text as if it were rendered.

    text
        Any string.
    font_size
        Integer indicating desired font size.
    font_style
        String. Either 'default' for default font or
        'mono' for monospace font.
    padding
        The horizontal and vertical padding for the text
        in pixels.
    """
    font = get_font(font_size, font_style)
    width, height = font.size(text)
    increment = 2 * padding
    padded_size = width + increment, height + increment
    return padded_size


def fit_text(text, font, max_width, ommit_direction):
    """Return optimal text to fit max_width passed."""
    ### if text as it fits width, then return earlier
    if not font.size(text)[0] > max_width:
        return text

    if ommit_direction == "left":
        slice_obj = slice(1, None)
    else:
        slice_obj = slice(None, -1)

    for _ in range(len(text)):
        text = text[slice_obj]
        if not font.size(text)[0] > max_width:
            break

    if ommit_direction == "left":
        text = "..." + text[3:]
    else:
        text = text[:-3] + "..."

    return text


def render_text(
    text,
    font_size=22,
    foreground_color=BLACK,
    background_color=(*BLACK, 0),
    antialiased=True,
    font_style="default",
    padding=5,
    max_width=0,
    ommit_direction="right",
    return_obj=False,
    coordinates_name="topleft",
    coordinates_value=(0, 0),
):
    """Return surface or object representing rendered text.

    text
        Any string.
    font_size
        Integer indicating desired font size.
    foreground_color
    background_color
        A tuple or list of r, g, b values which are
        integers ranging from 0 to 255. The background
        color can be None, though, in which case the
        background is rendered transparent. For the
        background color, an optional fourth value can
        be passed, which is an integer in the same range
        as the others, representing the image opacity (0 for
        full transparency and 255 for full opacity).
    antialiased
        Boolean indicating whether or not the text on
        the surface should be antialiased.
    font_style
        String. Either 'default' for default font or
        'mono' for monospace font.
    padding
        The horizontal and vertical padding for the text
        in pixels.
    max_width
        If provided, the text surf won't assume a width
        larger than max width.
    ommit_direction
        Part of the text to ommit in case its width surpass
        the max_width (if applicable).
    return_obj
        Boolean indicating whether or not a
        appcommon.autoblit.BasicObject instance should be
        returned. Default is False, and causes a
        pygame.Surface instance to be returned instead.
        In case this parameter is True, the coordinates_name
        and coordinates_value parameters are used.
    coordinates_name
        String representing attribute name of rect wherein
        to store the position information from the
        coordinates value parameter.
    coordinates_value
        Position information in the form of a tuple or list
        containing 2 integers representing positions
        in the x and y axes, respectively.
    """
    font = get_font(font_size, font_style)

    if max_width:
        max_width = abs(max_width)
        text = fit_text(text, font, max_width, ommit_direction)

    ### define background alpha for next operations

    try:
        has_transparency = background_color[3] < 255
    except IndexError:
        has_transparency = False

    ### render and adjust text surface according to
    ### background alpha

    text_surf = font.render(text, antialiased, foreground_color).convert_alpha()

    if has_transparency:

        bg = Surface(text_surf.get_size()).convert_alpha()
        bg.fill(background_color)
        bg.blit(text_surf, (0, 0))
        surf = bg

    else:
        surf = font.render(
            text, antialiased, foreground_color, background_color
        ).convert()

    ### apply padding if requested

    if padding:

        padding = abs(padding)

        ## calculate final width
        width, height = surf.get_size()
        new_width, new_height = (width + (2 * padding), height + (2 * padding))

        ## create new background surface with appropriate
        ## color according to presence of transparency
        if has_transparency:
            new_bg = Surface((new_width, new_height)).convert_alpha()

        else:
            new_bg = Surface((new_width, new_height)).convert()

        new_bg.fill(background_color)

        ## blit text on new background and reference as
        ## surf surf

        new_bg.blit(text_surf, (padding, padding))
        surf = new_bg

    if return_obj:
        obj = BasicObject()
        obj.image = surf

        obj.rect = obj.image.get_rect()
        setattr(obj.rect, coordinates_name, coordinates_value)

        return obj

    else:
        return surf


def get_str_slices(text, font_size, max_width):
    """Return list of string complying with maximum width.

    The str provided is divided into slices which once
    rendered into text surfaces will comply with the
    maximum width requirement. The str slices are word
    wrapped.

    text
        Any string.
    font_size
        Integer indicating desired font size.
    max_width
        Integer representing maximum width in pixels the
        text surfaces would have if str slices were
        rendered.
    """
    ### create container and alias
    str_slices = []

    ### iterate over text, trimming out words until the
    ### remaining text fits the maximum width and is stored
    ### in str_slices list, then do the same with the
    ### trimmed out text until nothing is left.

    trimmed = ""
    while text:
        width, _ = get_text_size(text, font_size)

        ## if don't fit max width, trim text
        if width > max_width:

            ## trim last word from text and stack on 'trimmed'

            lsi = text.rfind(" ")  # lsi = last_space_index
            text, trimmed = text[:lsi], text[lsi:] + trimmed

        ## otherwise store text lstripped text
        ## and get trimmed text, if any;
        ## (when trimmed is empty, the loop will exit
        ## automatically)
        else:
            str_slices.append(text.lstrip())
            text, trimmed = trimmed, ""

    return str_slices



"""Facility for dialogue generation."""

from math import pi

### third-party imports

from pygame import Rect
from pygame.draw import rect as draw_rect
from pygame.draw import circle as draw_circle
from pygame.draw import line as draw_line
from pygame.draw import arc as draw_arc
from pygame.draw import polygon as draw_polygon

### local imports

from ..palette import BLACK, ARBITRARY_COLORKEY, DIALOGUEBOX_BACKGROUND, has_same_rgb
from .surf import render_rect
from .text.main import render_text


# XXX
# Make the box adjust height too,
# so you're able to change font size
# and the box would adjust  accordingly.
# Use a fixture for that. Then maybe
# feature toggle.


def dialoguebox_factory(
    dialogue, object_pos=None, background_color=DIALOGUEBOX_BACKGROUND
):
    """Return a list of interlocutor/surface pairs.

    ### ASCII art: generated surfaces take this final form:
     _______________     ____________________
    (               )   (                    )
    | Hello, world! |   | Hello, world! Bye! |
    (______  _______)   (_________  _________)
           \/                     \/

    Notice the length is automatically adjusted to
    comprise the text surface.

    dialogue
        A list of dicts. Each dict contains a single field
        which specifies the interlocutor and whose values
        consists of its spoken line.

    object_pos (defaults to None)
        Position of interlocutor when it isn't the pc.

    background_color
        A tuple or list of r, g, b values which are
        integers ranging from 0 to 255. This color will
        be in several pygame.draw drawing calls.

        Though this function wasn't intended to be used
        in this way an optional alpha value within the
        same range can also be provided in the tuple
        or list, and the purpose of this can be read
        in the link bellow:
        https://www.pygame.org/docs/ref/draw.html

        Here's the specific excerpt:
        "Most of the arguments accept a color argument
        that is an RGB triplet. These can also accept
        an RGBA quadruplet. The alpha value will be
        written directly into the Surface if it contains
        pixel alphas, but the draw function will not draw
        transparently. The color argument can also be an
        integer pixel value that is already mapped to the
        Surface's pixel format."
    """
    if object_pos is None:
        for line in dialogue:
            for key in line:
                if key != "player":
                    msg = (
                        "object_pos must be provided "
                        + "an interlocutor other than "
                        + "player speaks."
                    )
                    raise ValueError(msg)

    if has_same_rgb(background_color, ARBITRARY_COLORKEY):
        msg = "'background_color' can't be equal to" + " 'ARBITRARY_COLORKEY' constant."
        raise ValueError(msg)

    bg_color = background_color
    common_width = 4

    dialogue_boxes = []
    for textline in dialogue:
        ### Initial surf to be used as a canvas
        interlocutor, line_str = textline.copy().popitem()

        if interlocutor == "self":
            interlocutor_pos = object_pos
        else:
            interlocutor_pos = "player_pos"

        text_surf = render_text(
            line_str, font_size=22, foreground_color=BLACK, padding=0
        )

        box_width = text_surf.get_size()[0] + 40

        surf = render_rect(box_width, 65)
        surf.fill(ARBITRARY_COLORKEY)

        ### Dialoguebox volume (blocking)

        ## 02 rects
        #
        #    rect1
        # |---------------|
        #     _________      _
        #  __|         |__   |
        # |__           __|  |  rect2
        #    |_________|     |
        #
        #
        rect1 = Rect((0, 20), (box_width, 5))
        rect2 = Rect((20, 0), (box_width - 40, 45))
        draw_rect(surf, bg_color, rect1)
        draw_rect(surf, bg_color, rect2)

        ## 04 circles
        #
        # tlpos _________ trpos
        #    __.         .__
        #   |__.         .__|
        # blpos|_________| brpos
        #
        # each point above represents the center where
        # the circles will be drawn. This way it'll appear
        # that the box has round edges.
        #
        # Also, in next code blocks,
        # tl means topleft, tr means topright,
        # bl means bottomleft and  br means bottomright.
        tlpos, trpos, blpos, brpos = (
            (20, 20),
            (box_width - 20, 20),
            (20, 25),
            (box_width - 20, 25),
        )
        radius = 20
        draw_circle(surf, bg_color, tlpos, radius)
        draw_circle(surf, bg_color, trpos, radius)
        draw_circle(surf, bg_color, blpos, radius)
        draw_circle(surf, bg_color, brpos, radius)

        ### Dialogue box outlines and final touches

        # 04 arc
        # will be drawn around the round corners of the
        # volume
        arc_rect1 = Rect((0, 0), (40, 40))
        arc_rect2 = Rect((0, 8), (40, 40))
        arc_rect3 = Rect((box_width - 41, 0), (40, 40))
        arc_rect4 = Rect((box_width - 40, 8), (40, 40))

        startangle1, startangle2, startangle3, startangle4 = pi / 2, pi, 0, (3 * pi) / 2

        endangle1, endangle2, endangle3, endangle4 = pi, (3 * pi) / 2, pi / 2, 0

        draw_arc(surf, BLACK, arc_rect1, startangle1, endangle1, common_width)
        draw_arc(surf, BLACK, arc_rect2, startangle2, endangle2, common_width)
        draw_arc(surf, BLACK, arc_rect3, startangle3, endangle3, common_width)
        draw_arc(surf, BLACK, arc_rect4, startangle4, endangle4, common_width)

        # eliminating artifacts in the drawn arcs
        square = Rect((box_width - 22, 48), (4, 4))
        draw_rect(surf, ARBITRARY_COLORKEY, square)

        for i in range(3):
            arc_rect1.y += 1
            draw_arc(surf, BLACK, arc_rect1, startangle1, endangle1, 1)
        for i in range(2):
            arc_rect1.x += 1
            draw_arc(surf, BLACK, arc_rect1, startangle1, endangle1, 1)

        for i in range(3):
            arc_rect2.y += -1
            draw_arc(surf, BLACK, arc_rect2, startangle2, endangle2, 1)
        for i in range(2):
            arc_rect2.x += 1
            draw_arc(surf, BLACK, arc_rect2, startangle2, endangle2, 1)

        for i in range(3):
            arc_rect3.y += 1
            draw_arc(surf, BLACK, arc_rect3, startangle3, endangle3, 1)
        for i in range(2):
            arc_rect3.x += -1
            draw_arc(surf, BLACK, arc_rect3, startangle3, endangle3, 1)
        arc_rect3.x += 3
        draw_arc(surf, BLACK, arc_rect3, startangle3, endangle3, 1)
        arc_rect3.y -= 3
        draw_arc(surf, BLACK, arc_rect3, startangle3, endangle3, 1)

        for i in range(3):
            arc_rect4.y += -1
            draw_arc(surf, BLACK, arc_rect4, startangle4, endangle4, 1)
        for i in range(2):
            arc_rect4.x += -1
            draw_arc(surf, BLACK, arc_rect4, startangle4, endangle4, 1)

        # 04 lines
        # will be drawn to outline the straight sections
        # of the volume
        #
        # in the end, the box will be more or less like
        # this:
        #  _______________
        # (               )
        # |               |
        # (_______________)
        #
        start1, end1 = (20, 1), (box_width - 20, 1)
        start2, end2 = (19, 45), (box_width - 20, 45)
        start3, end3 = (1, 19), (1, 27)
        start4, end4 = (box_width - 3, 20), (box_width - 3, 27)
        draw_line(surf, BLACK, start1, end1, common_width)
        draw_line(surf, BLACK, start2, end2, common_width)
        draw_line(surf, BLACK, start3, end3, common_width)
        draw_line(surf, BLACK, start4, end4, common_width)

        # triangle
        #  _______________
        # (               )
        # |               |
        # (______  _______)
        #        \/ <--this triangle here
        #
        # It will be drawn over the outline of the volume
        # so that it appear to be part of the box.
        mid = round(box_width / 2)
        pointlist = ((mid, 64), (mid - 20, 44), (mid + 20, 44))
        draw_polygon(surf, bg_color, pointlist)

        # then we outline the part of the triangle
        # outside the volume
        start1, end1 = (mid - 20, 44), (mid, 64)
        start2, end2 = (mid + 20, 44), (mid, 64)
        draw_line(surf, BLACK, start1, end1, common_width)
        draw_line(surf, BLACK, start2, end2, common_width)

        ### Blitting the text and making other adjustments

        surf.blit(text_surf, (19, 8))

        # Making everything else outside box transparent
        surf.set_colorkey(ARBITRARY_COLORKEY)

        dialogue_pair = (interlocutor_pos, surf)
        dialogue_boxes.append(dialogue_pair)

    return dialogue_boxes


# XXX
# Ideas for improvement:
# - I don't know if the circle drawing algorithm generate the
#   same noise the arcs generate when using thick lines, if
#   if they don't, then arcs could be produced from circles
#   with some parts hidden by 'colorkeyed' areas.



"""Facility for displaying and managing prompts."""

### local imports

from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.surf import INVISIBLE_SURF


class PromptManager:
    """Display and manages prompt boxes."""

    def __init__(self, player):
        """Initialize superclass and set variables.

        player
            A reference to a player.main.Player instance.
        """
        self.player = player

        self.image = INVISIBLE_SURF
        self.rect = self.image.get_rect()

        self.prompt_display = None

        self.update = empty_function

    def draw(self):
        """Draw prompt display if available."""
        try:
            self.prompt_display.draw()

        ### attribute error due to prompt_display being
        ### None and this object lacking a draw method
        except AttributeError:
            pass

    def get_prompt(self, prompt):
        """Instantiate a prompt box.

        prompt
            A tuple containing a pair of data. The
            first item includes the position where a surface
            should appear and said surface. The second one
            is an action to be executed should the player
            confirm the prompt.
        """
        prompt_pair = prompt[0]
        prompt_action = prompt[1]
        self.prompt_display = PromptDisplay(
            self, self.player, prompt_pair, prompt_action
        )

    def finish_prompt(self):
        """Set current prompt to None."""
        self.prompt_display = None

    def confirm(self):
        """Execute prompt action."""
        if self.prompt_display:
            self.prompt_display.confirm()
            self.prompt_display = None


class PromptDisplay(BasicObject):
    """Object for prompt."""

    def __init__(self, manager, player, prompt_pair, prompt_action):
        """Initialize superclass and prepare sprites.

        manager
            The instance of the player.PromptManager class.
        player
            A reference to a player.main.Player instance.
        prompt_pair
            A tuple containing a pair of data. The
            first item includes the position where a surface
            should appear and the second one said surface.
        prompt_action
            A callable to be executed should the player
            confirm the prompt.
        """
        self.manager = manager
        self.player = player

        self.image = prompt_pair[1]

        self.rect = self.image.get_rect()
        self.rect.midbottom = prompt_pair[0]

        self.confirm = prompt_action

        self.update = empty_function



"""A head sprite to represent the pc."""

### local imports

from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.surf import render_image


class Head(BasicObject):
    """A player's head featured alongside healthbar."""

    def __init__(self):
        """Initialize superclass and set variables."""
        self.image = render_image("dani_head.png")

        self.rect = self.image.get_rect()
        self.rect.topleft = (30, 30)

        self.update = empty_function



"""Facility for escape menu class and related tools."""

from collections import deque
from functools import partialmethod

### third-party imports

from pygame import QUIT, KEYDOWN, K_RETURN, K_ESCAPE, K_w, K_s

from pygame.display import update
from pygame.event import get as get_events

### local imports

from ..config import SCREEN, GAME_REFS

from ..appcommon.surf import render_rect
from ..appcommon.autoblit import BlitterSet, BasicObject
from ..appcommon.exception import QuitGameException, ManagerSwitchException
from ..appcommon.text.main import render_text
from ..appcommon.behaviour.buffer import remove_buffers

from .classes import remove_level_reference

from ..palette import BLACK, WHITE


class EscapeMenu(object):
    """Presents an escape menu."""

    def __init__(self, level):
        """Initialize instance.

        level
            the instance of gamelevel.main.Level."""
        self.font_size = 32
        self.level = level

        self.widget_names = ["Resume game", "Quit to main menu", "Quit to desktop"]

        self.background = render_rect(400, 420)
        self.build_items()
        self.set_arrow_widget()

    def build_items(self):
        """Create menu item widgets."""
        self.menu_widgets = deque()
        x, y = 620, 320
        for name in self.widget_names:
            escape_menu_widget = EscapeMenuWidget(name, self, self.level, (x, y))
            self.menu_widgets.append(escape_menu_widget)
            y += 50

        self.widgets_group = BlitterSet(self.menu_widgets)

        self.selected_item = self.menu_widgets[0]
        self.widgets_group.update(self.menu_widgets)

    def set_arrow_widget(self):
        """Create a highlight arrow for selected items."""
        self.arrow_widget = BasicObject()
        self.arrow_widget.image = render_text(
            ">",
            font_size=self.font_size,
            foreground_color=WHITE,
            background_color=BLACK,
        )
        self.arrow_widget.rect = self.arrow_widget.image.get_rect()
        self.arrow_widget.rect.midright = self.selected_item.rect.midleft
        self.widgets_group.add(self.arrow_widget)

    def control(self):
        """Handle events in level event queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            elif event.type == KEYDOWN:

                if event.key == K_w:
                    self.select_previous()

                elif event.key == K_s:
                    self.select_next()

                elif event.key == K_RETURN:
                    self.invoke_item_action()

                elif event.key == K_ESCAPE:
                    raise ManagerSwitchException(self.level)

    def update(self):
        """Update menu state."""
        self.arrow_widget.rect.midright = self.selected_item.rect.midleft

    def draw(self):
        """Draw objects."""
        SCREEN.blit(self.background, (580, 260))
        self.widgets_group.draw()

        update()

    def rotate_selection(self, amount):
        """Assign selection after rotating by amount."""
        self.menu_widgets.rotate(amount)
        self.selected_item = self.menu_widgets[0]

    select_previous = partialmethod(rotate_selection, 1)
    select_next = partialmethod(rotate_selection, -1)

    def invoke_item_action(self):
        """Invoke the action on selected menu item."""
        self.selected_item.action()


class EscapeMenuWidget(object):
    """Escape menu item widget."""

    def __init__(self, name, escape_menu, level, topleft):
        """Initialize superclass and set instance.

        name
            string representing the item text. Also used
            to associate it with its method.
        escape_menu
            the instance of ingame.escapemenu.EscapeMenu
            class.
        level
            the instance of gamelevel.main.Level
        topleft
            the topleft coordinates for the rect attribute
        """
        self.font_size = 32

        self.name = name
        self.escape_menu = escape_menu
        self.level = level

        if name == "Resume game":
            self.action = self.resume_game
        elif name == "Quit to main menu":
            self.action = self.quit_to_main_menu
        elif name == "Quit to desktop":
            self.action = self.exit_game

        self.image = render_text(
            name,
            font_size=self.font_size,
            foreground_color=WHITE,
            background_color=BLACK,
        )
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

    def resume_game(self):
        """Resume game."""
        raise ManagerSwitchException(self.level)

    def quit_to_main_menu(self):
        """Load main menu."""
        remove_level_reference()

        remove_buffers()

        del GAME_REFS.dirname

        raise ManagerSwitchException("main_menu")

    def exit_game(self):
        """Exit game."""
        raise QuitGameException



"""Facility for parallax mountains."""

### local imports

from ..config import SCREEN_RECT
from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.surf import render_image


# XXX
# maybe smooth the edges of the mountain.png image.


class Mountain(BasicObject):
    """Base class for mountains."""

    def __init__(self):
        """Position and perform setups.

        bottomleft
            Bottomleft coordinates for a pygame.Rect instance.
        """
        self.image = render_image("mountain.png")
        self.rect = self.image.get_rect()
        self.rect.midbottom = SCREEN_RECT.midbottom

        # 40: just an arbitrary value.
        # It means we can scroll down the screen
        # 10 times and the image still won't leave the
        # screen boundaries.
        self.rect.y += 40

        self.update = empty_function

    def scroll(self, dx, dy):
        """Scroll by dx and dy in respective axis.

        dx
            Integer indicating scrolling in x axis.
        dy
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(dx / 160, dy / 180)



"""Facility for 2d point generation for general purpose."""

# XXX
# (see edit at the end)
#
# This calculations/generations/changes have many
# different applications. I could probably make a game
# just out of the exploration of them. I should stop by
# now, but keep making notes about possibilities for
# future implementations:
#
# Spring like movement:
# Moving like a spring should be achieved by simulating
# a body 'orbiting' around a second body while this
# second body is in movement (either circular or in a
# specific direction). That is, translating the first
# body circular movement by the movement of the second.
# Also, such circular movement could also be elliptical
# instead.
#
# Spiral like movement:
# Moving in spirals should be achieved by making lots of
# circles of different radius and making a sequence of
# the first index of the innermost circle, then the second
# index of the second circle, then third index of third
# circle and so on.
#
# Spiral and spring movements might be combinable, I think,
# to simulate for instance how some fireworks spiral away
# (dangerously so); it could be useful for sparks, fireworks,
# etc.
#
# Wave movement:
# an object going up and down following another one going
# straight. If up and down speed is constant, we have a
# zig-zag:
#
# .       .       .
#  .     . .     . .
#   .   .   .   .   .
#    . .     . .     .
#     .       .       .
#
# Also, instead of translating the points per se, you could
# just do loops and parent the object to another object
# doing a different loop or moving along, which would then
# cause the other object to be translated along.
#
# All very interesting applications.
#
# edit: this should become a docstring

from math import pi, sin, cos, hypot
from collections import deque
from collections.abc import Sequence
from random import seed, shuffle, sample
from itertools import cycle
from copy import deepcopy

### local imports

from .math import invert_point, offset_point
from .indexgen import generate_indices

# XXX
# idea: use rotation matrix formula to rotate points
# around a given origin (defaulting to [0, 0])

# XXX
# idea: obtaining points from filtering existing ones
# for instance, obtaining a half-circle from filtering
# point where y > 0, etc.


### utility function (use elsewhere, may also be used here
### if needed)


def convert_items_to_tuples(point_sequence):
    """Convert items in a sequence to tuples.

    Used to convert sequences whose items represent 2d points
    into tuples. This function was created because
    pygame.Rect updates its position much quicker when
    using tuples in comparison with other types (this was
    properly profiled using standard library timeit module).
    """
    for i in range(len(point_sequence)):
        point_sequence[i] = tuple(point_sequence[i])


### main function


def generate_points(pos_data, pos_map={}):
    """Return points list after generating/changing them.

    pos_data (dict or list)
        Data used to retrieve/load/generate/alter
        data recursively or not.
    pos_map (dict)
        contains list of positions (2d points/vectors).
        It is only meant to provide values and never be
        altered here.
    """

    ### if data provided is list, then you don't need to
    ### process anything; just return it
    if isinstance(pos_data, list):
        return pos_data

    ### process operation to
    ### retrieve/load/generate/alter points
    name = pos_data["name"]

    ## retrieve point list from pos_map
    if name == "retrieve_from_map":
        key = pos_data["key"]
        output = pos_map[key]

    ## generate point list for a line
    elif name == "line":
        quantity = pos_data["how_many_points"]
        start_point = pos_data.get("start_point", (0, 0))
        end_point = pos_data["end_point"]

        distance_x = end_point[0] - start_point[0]
        distance_y = end_point[1] - start_point[1]

        # compesate range starting in 1
        range_offset = 1
        # discount presence of start and end point
        existing_discount = -2

        offset_quantity = quantity + range_offset + existing_discount

        x_step = distance_x / offset_quantity
        y_step = distance_y / offset_quantity

        output = []
        output.append(start_point)

        for i in range(1, offset_quantity):
            amount_x = i * x_step
            amount_y = i * y_step
            x = round(start_point[0] + amount_x)
            y = round(start_point[1] + amount_y)
            output.append((x, y))

        output.append(end_point)

    ## generate point list for a circle
    elif name == "circle":

        quantity = pos_data["how_many_points"]
        radius = pos_data["radius"]

        output = []
        for k in range(quantity):
            value = (k * 2 * pi) / quantity
            x = round(radius * cos(value))
            y = round(radius * sin(value))
            point = (x, y)
            output.append(point)

    ## generate point list by mapping sequence to another
    ## (translating points of 1st one according to 2nd one)
    elif name == "sequence_to_path":
        ### process sequence content
        sequence = generate_points(pos_data["sequence"], pos_map)

        ### process path content
        path = generate_points(pos_data["path"], pos_map)

        ### generate points

        output = []
        cyclic_sequence = cycle(sequence)

        for point in path:
            translated_point = generate_points(
                {
                    "name": "translate",
                    "sequence": [next(cyclic_sequence)],
                    "sequence2": [point],
                }
            ).pop()

            output.append(translated_point)

    ## generate point list by repeating points
    elif name == "repeat":
        how_many_times = pos_data["how_many_times"]
        sequence = generate_points(pos_data["sequence"], pos_map)

        output = sequence * how_many_times

    ## generate point list by simulating accel. movement
    elif name == "from_accelerated_movement":
        start_point = pos_data.get("start_point", [0, 0])
        speed = pos_data.get("speed", [0, 0])
        acceleration = pos_data.get("acceleration", [0, 0])
        quantity = pos_data.get("how_many_points", 0)

        x, y = start_point
        speed_x, speed_y = speed
        accel_x, accel_y = acceleration

        output = []
        for i in range(quantity):
            x += speed_x
            y += speed_y

            point = (x, y)
            output.append(point)

            speed_x += accel_x
            speed_y += accel_y

    ## generate point list by movement simulation
    elif name == "spiral":
        quantity = pos_data["how_many_points"]
        resolution = pos_data["resolution"]
        radius_incr = pos_data["radius_increment"]

        cyclic_indices = cycle(range(resolution))

        output = []
        for i in range(quantity):
            radius = i * radius_incr
            circle = generate_points(
                {"name": "circle", "how_many_points": resolution, "radius": radius},
                pos_map,
            )
            point = circle[next(cyclic_indices)]
            output.append(point)

    ## generate point list by zipping two indice lists
    # XXX
    # more possibilities:
    # permutation and combination with itertools
    elif name == "zip_sequences":
        indices_x = generate_indices(pos_data["indices_x"])
        indices_y = generate_indices(pos_data["indices_y"])

        output = [(x, y) for x, y in zip(indices_x, indices_y)]

    ## generate points from cartesian product of indices
    elif name == "product":
        indices = generate_indices(pos_data["indices"])
        try:
            indices2 = pos_data["indices2"]
        except KeyError:
            indices2 = indices
        else:
            indices2 = generate_indices(indices2)

        output = [(x, y) for x in indices for y in indices2]

    ## generate random point list from points inside rect
    elif name == "random_from_rect":

        rect_size = pos_data["size"]
        quantity = pos_data["how_many_points"]

        width, height = rect_size

        population = [(x, y) for x in range(width) for y in range(height)]

        seed()
        output = sample(population, quantity)

    ## generate random point list from points inside circle
    elif name == "random_from_circle":
        radius = pos_data["radius"]
        quantity = pos_data["how_many_points"]

        population = [
            (x, y)
            for x in range(-radius, radius + 1)
            for y in range(-radius, radius + 1)
            if hypot(x, y) <= radius
        ]

        seed()
        output = sample(population, quantity)

    ## deepcopy a list of positions
    elif name == "deepcopy":
        sequence = generate_points(pos_data["sequence"], pos_map)

        output = deepcopy(sequence)

    ## change point list by mirroring
    elif name == "mirror":
        sequence = generate_points(pos_data["sequence"], pos_map)

        use_last = pos_data.get("use_last", True)

        mirror_x = pos_data.get("mirror_x", True)
        mirror_y = pos_data.get("mirror_y", True)

        ### either use last or first point
        if use_last:
            mirror_point = sequence[-1]
        else:
            mirror_point = sequence[0]
        m_point_x, m_point_y = mirror_point

        ### sort temp list according to 'use_last'
        iterable = sequence[::-1] if use_last else sequence

        ### generate new points
        output = []
        for point in iterable:
            x, y = point
            if x == m_point_x and (y == m_point_y):
                # ignore if it is the mirror point
                continue
            else:
                # increments are either zero or
                # double inverted distance
                # depending on axes set to be mirrored
                x_incr = 2 * (m_point_x - x) if mirror_x else 0

                y_incr = 2 * (m_point_y - y) if mirror_y else 0

                new_x = x + x_incr
                new_y = y + y_incr
                new_point = (new_x, new_y)
                output.append(new_point)

        ### append or prepend according to 'use_last'
        if use_last:
            sequence.extend(output)
            output = sequence
        else:
            output.reverse()
            output.extend(points)

    ## change point list by repeating it
    elif name == "repeat_existing":
        sequence = generate_points(pos_data["sequence"], pos_map)
        how_many_times = pos_data["how_many_times"]
        output = sequence * how_many_times

    ## change point list by executing pre-defined operation
    elif name == "back_and_forth_from_existing":
        sequence = generate_points(pos_data["sequence"], pos_map)
        first_half = sequence[:]
        last_half = sequence[::-1][1:-1]

        output = first_half + last_half

    ## change point list by multiplying existing ones
    elif name == "multiply":
        sequence = generate_points(pos_data["sequence"], pos_map)

        multiplier = pos_data.get("multiplier", False)

        output = []
        if multiplier:
            ## multiply all points by multiplier
            for item in sequence:
                seq = [
                    item,
                ] * multiplier
                output.extend(seq)

        else:
            ## else multiply points according to map

            # for simplification, instead of multiplicands,
            # the fields are indices representing point
            # position in sequence
            multiplier_map = pos_data["multiplier_map"]
            for index, item in enumerate(sequence):
                multiplier = multiplier_map.get(str(index), 1)
                seq = [
                    item,
                ] * multiplier

                output.extend(seq)

    ## change point list by slicing it
    elif name == "slice":
        sequence = generate_points(pos_data["sequence"], pos_map)
        start = pos_data.get("start")
        stop = pos_data.get("stop")
        step = pos_data.get("step")

        output = sequence[start:stop:step]

    ## change point list by inverting them
    elif name == "invert":
        sequence = generate_points(pos_data["sequence"], pos_map)

        invert_x = pos_data.get("invert_x", False)
        invert_y = pos_data.get("invert_y", False)

        # XXX the comprehension below produces a list of
        # tuples (which represent points); should we
        # convert such points into lists?

        output = [invert_point(point, invert_x, invert_y) for point in sequence]

    ## change point list by translating it
    elif name == "translate":

        sequence = generate_points(pos_data["sequence"], pos_map)

        try:
            sequence2 = pos_data["sequence2"]

        except KeyError:
            delta_x = pos_data.get("delta_x", 0)
            delta_y = pos_data.get("delta_y", 0)

            offset = (delta_x, delta_y)

            translated_points = []
            for point in sequence:
                new_point = offset_point(point, offset)
                translated_points.append(new_point)

        else:

            sequence2 = generate_indices(sequence2)
            translated_points = []

            for point, offset in zip(sequence, sequence2):
                new_point = offset_point(point, offset)
                translated_points.append(new_point)

        output = translated_points

    ## change point list by deque rotating existing ones
    elif name == "deque_rotate":
        sequence = generate_points(pos_data["sequence"], pos_map)

        rotation_amount = pos_data["rotation_amount"]

        points_deque = deque(sequence)
        points_deque.rotate(rotation_amount)

        output = list(points_deque)

    ## change point list by shuffling them
    elif name == "shuffle":
        sequence = generate_points(pos_data["sequence"], pos_map)
        seed()
        shuffle(sequence)

        output = sequence

    ### make sure return value is a sequence

    if isinstance(output, Sequence):
        return output

    else:
        return [output]



"""Game classes management for interactive levels.

Interactive levels are instances of
gamelevel.main.Level which is located here:
gamelevel/main.py
"""

from ..backprop.backprops import StaticBackProp

from ..block.blocks import StaticBlock
from ..block.grassdirt import GrassDirtBlock

from ..middleprop.sign import Sign
from ..middleprop.middleprops import VinesLadder, SimpleDoor, CaveEntrance
from ..middleprop.mushroomplant import MushroomPlant
from ..middleprop.satellitedish import SatelliteDish

from ..actor.npc import NPC
from ..actor.rocketscorpion import RocketScorpion
from ..actor.thornshooter.main import ThornShooter

from ..frontprop.frontprops import LeafCarpet, FrozenCover
from ..frontprop.toxicpollencloud import ToxicPollenCloud


CLASS_MAP = {
    "dirt_back_prop": StaticBackProp,
    "dirt_darker_back_prop": StaticBackProp,
    "stone_darker_back_prop": StaticBackProp,
    "sand_back_prop": StaticBackProp,
    "stone_back_prop": StaticBackProp,
    "grass_and_dirt_block": GrassDirtBlock,
    "stone_block": StaticBlock,
    "dirt_block01": StaticBlock,
    "sand_block": StaticBlock,
    "simple_door": SimpleDoor,
    "cave_entrance": CaveEntrance,
    "mushroom_plant": MushroomPlant,
    "vines_climbable": VinesLadder,
    "vines_top": VinesLadder,
    "vines_grid_climbable": VinesLadder,
    "vines_grid_top": VinesLadder,
    "sign": Sign,
    "satellite_dish": SatelliteDish,
    "leaf_carpet": LeafCarpet,
    "toxic_pollen_cloud": ToxicPollenCloud,
    "frozen_cover": FrozenCover,
    "thorn_shooter": ThornShooter,
    "rocket_scorpion": RocketScorpion,
    "npc": NPC,
}

LAYER_MAP = {
    "dirt_back_prop": "back_props",
    "dirt_darker_back_prop": "back_props",
    "stone_darker_back_prop": "back_props",
    "sand_back_prop": "back_props",
    "stone_back_prop": "back_props",
    "cave_entrance": "middle_props",
    "simple_door": "middle_props",
    "mushroom_plant": "middle_props",
    "vines_climbable": "middle_props",
    "vines_top": "middle_props",
    "vines_grid_climbable": "middle_props",
    "vines_grid_top": "middle_props",
    "sign": "middle_props",
    "satellite_dish": "middle_props",
    "dirt_block01": "blocks",
    "grass_and_dirt_block": "blocks",
    "stone_block": "blocks",
    "sand_block": "blocks",
    "thorn_shooter": "actors",
    "rocket_scorpion": "actors",
    "npc": "actors",
    "leaf_carpet": "front_props",
    "toxic_pollen_cloud": "front_props",
    "frozen_cover": "front_props",
}


def insert_level_reference(level):
    """Give a level reference for each class.

    level
        An instance of gamelevel.main.Level class.
    """
    for class_key in CLASS_MAP:
        CLASS_MAP[class_key].level = level


def remove_level_reference():
    """Remove level reference for each class.

    This is done by assigning None to it.
    """
    for class_key in CLASS_MAP:
        CLASS_MAP[class_key].level = None


def get_instance(data):
    """Return game class instance according to data.

    data
        A dict with a prop_name field which tells which
        kind of object to instantiate."""
    prop_name = data["prop_name"]
    Instantiator = CLASS_MAP.get(prop_name, "dummy")
    instance = Instantiator(**data)

    return instance


def get_layer_name(prop_name):
    """Return level layer name according to provided data.

    prop_name
        A str telling the name of the prop. It is used by
        a dict to retrieve the name of layer wherein to
        put the prop.
    """
    layer_name = LAYER_MAP.get(prop_name, "dummy_layer")

    return layer_name



"""Documentation and tests for the Content View Class.

View of walking deque content w/ autonomous walking.

### Preliminary imports

Since some of our tests/examples use a standard library
import, let's go ahead and perform it:

standard library import
>>> from copy import deepcopy

let's also perform local imports
>>> from common.wdeque.main import WalkingDeque

### Introduction

WalkingDeque instances can generate instances of this
ContentView class. Instances of such class work like
read-only copies/proxies of the walking deques which
originated them.

Their rotations/walking state is totally independent, though.
The only difference is that when the original walking deques
have the value of their items reassigned, they also
automatically update the values of the corresponding items in
the content views.

Notice below, that the original walking deque and its content
view have independent walking (walking one doesn't change the
other):

>>> orig = WalkingDeque(range(4))
>>> view = orig.get_content_view()

>>> orig.walk(3)
>>> view.walk(1)

>>> orig
WalkingDeque([3, 0, 1, 2])
>>> view
ContentView([1, 2, 3, 0])

However, as stated, when we update the value of one of the
original deque items, the content view is also updated
accordingly:

>>> orig[0] = 45
>>> orig
WalkingDeque([45, 0, 1, 2])
>>> view
ContentView([1, 2, 45, 0])

Notice that though we updated the index 0 of the original
deque, it was the index 2 of the view that got updated. Why
is that? This is because the update is performed on the
corresponding original index of both deques. Here's the
proof:

>>> orig.get_original_index(0)
3
>>> view.get_original_index(2)
3

That is, both the current value of index 0 of the original
deque and the value of index 2 of the content view were
originally in the same original position when they were
instantiated, see:

>>> orig.get_unwalked_list() == view.get_unwalked_list()
True


### Item reassignment for content views

Since they are read-only objects, content views can't have
their items reassigned. Their contents depend only on the
contents of the "proxied" deque.

>>> view[0] = 47
Traceback (most recent call last):
...
ValueError: value must be identical to proxied deque value

Trying to reassign a value manually on the content view
will work, though, on only condition: when the corresponding
value in the original deque is identical. For instance, since
the index 0 of the view has the same original index of the
current index 2 of the original deque, which has the value 1,
if you assign 1 again to the index 0 of the view no error
will be raised:

>>> orig
WalkingDeque([45, 0, 1, 2])
>>> view
ContentView([1, 2, 45, 0])
>>> view[0] = 1
>>> # as you can see, no error was raised

This is a seemingly useless operation, at least in this
example, since the value replaced remains the same, but this
is an essential mechanism which works in synergy with both
the WalkingDeque and ContentView classes to allow then to
have their proxy-view relationship.

In summary, the views are tightly protected from
indiscriminate item reassignment.

## Item reassignment and identity check

But notice that it's not enough for the items to have equal
values, they must also be --the same-- object, which means
they must pass an identity check (a is b):

>>> names = ['Anne', 'Mary']
>>> orig = WalkingDeque([0, 1, 2, names])
>>> view = orig.get_content_view()

>>> names_copy = names.copy()

>>> # the reassignment bellow will raise an error even though
>>> # the values are the same
>>> view[3] = names_copy
Traceback (most recent call last):
...
ValueError: value must be identical to proxied deque value

The error above happens because...
>>> orig[3] is not names_copy
True

However...
>>> orig[3] is names
True

...so the reassignment below works, it doesn't raise errors:
>>> view[3] = names


### "proxied" attribute

The content views store a reference to their proxied deque
in the proxied attribute:

>>> view.proxied is orig
True

Conversely, though this is a detail of the WalkingDeque
class, it is useful to know that views are also referenced
in their proxied deque "views" attribute, which holds a
list of references:

>>> orig.views[0] is view
True


### Sibling content views

The content views can also generate sibling content views,
meaning the content view generated are content views of the
proxied deque, not of view that seemed to generate it.

This is because the get_content_view method in this class
works by asking their proxied deque to generate the new
content view for them. See examples below (notice that
sometimes we perform the walk method on some views; this
isn't necessary, but we do so to remember you that the
walking of all objects are independent):

>>> orig = WalkingDeque([0, 1, 2, 3])
>>> view = orig.get_content_view()
>>> other_view = view.get_content_view()
>>> other_view.walk(2)
>>> other_view
ContentView([2, 3, 0, 1])

Notice that though other_view was generated from the
"view" get_content_view method, it isn't a content view of
"view", but of "orig".

>>> other_view.proxied is orig
True

Thus, view and other_view are both content views of the same
walking deque, "orig"

>>> view.proxied is other_view.proxied
True

Additionaly, walking deques maintain a list of their content
views in a views attribute:

>>> orig.views
[ContentView([0, 1, 2, 3]), ContentView([2, 3, 0, 1])]
>>> view is orig.views[0]
True
>>> other_view is orig.views[1]
True


### Custom deepcopy operation

The custom deepcopy operation in the ContentView class was
designed to work together with the custom deepcopy operation
in its proxied deque's class (WalkingDeque).

As long as they are deepcopied recursively (when they are
stored in the same object, either as items, keys or
attributes), the relationship between the deepcopies remains. 

Example of behaviour after recursive deepcopy:

>>> w  = WalkingDeque([0, 1, 2, 3])
>>> v1 = w.get_content_view()
>>> v2 = w.get_content_view()
>>> d = {
...   "original": w,
...   "view1"   : v1,
...   "view2"   : v2
... }
>>> new_dict = deepcopy(d)
>>> new_w    = new_dict["original"]
>>> new_v1   = new_dict["view1"]
>>> new_v2   = new_dict["view2"]
>>> new_v1.proxied is new_w
True
>>> new_v2.proxied is new_w
True
>>> new_w.walk(2)
>>> new_w[0] = 43
>>> new_v1[2] == 43
True

What happened above, in other words, is that since the
original deque and its views are part of the same object (the
dictionary), when such object was deepcopied, they were
deepcopied recursively, and the relationship between the
deepcopies new_w, new_v1 and new_v2 remains, even though they
are copies.

However, when the deepcopying isn't recursive (you deepcopy
in more than one step), the relationship between deepcopied
objects is lost:

>>> walking_deque = WalkingDeque([0, 1, 2, 3])
>>> content_view  = walking_deque.get_content_view()

>>> new_wd  = deepcopy(walking_deque)
>>> new_cv = deepcopy(content_view)

>>> # notice how the proxied deque of the new_cv content
>>> # view isn't new_wd:
>>> new_cv.proxied is new_wd
False
>>> new_wd.views[0] is new_cv
False

That is, though their original versions had a relationship,
they haven't preserved such relationship between themselves,
even though they are deepcopies of the same objects. Thus,
changes in the walking deque deepcopy won't affect the
view deepcopy.

>>> new_w.walk(2)
>>> new_w[0] = 22
>>> new_v1[2] == 22
False

So, if you want to preserve the relationship between
deepcopies, at least put the original objects into a tuple
before deepcopying them, for instance: 

>>> walking_deque = WalkingDeque([0, 1, 2, 3])
>>> content_view  = walking_deque.get_content_view()

>>> # deepcopy them inside a tuple:
>>> new_wd, new_cv = deepcopy((walking_deque, content_view))
>>> # and now the deepcopies have preserved their
>>> # relationship:
>>> new_cv.proxied is new_wd
True
>>> new_wd.views[0] is new_cv
True

As long as the objects deepcopied are embedded in the same
object or contained there somehow, the deepcopy will be
recursive. Unless of course, it is a custom class which have
some customized deepcopy operation.

"""



"""Facility for the front_props layer."""

### local imports

from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.surf import render_image


class LeafCarpet(BasicObject):
    """A leaf carpet on a wall."""

    level = None
    prop_map = {"leaf_carpet": render_image("leaf_carpet.png")}

    def __init__(self, prop_name, coordinates):
        """Initialize superclass and variables.

        prop_name
            string representing the name of the prop.
        coordinates
            the bottomleft coordinates for the rect attribute.
        """
        self.prop_name = prop_name

        self.image = LeafCarpet.prop_map[prop_name]

        self.rect = self.image.get_rect()

        self.rect.bottomleft = coordinates

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)


class FrozenCover(BasicObject):
    """A cover made of ice."""

    level = None
    prop_map = {"frozen_cover": render_image("frozen_cover.png")}

    def __init__(self, prop_name, coordinates):
        """Initialize superclass and variables.

        prop_name
            string representing the name of the prop.
        coordinates
            the bottomleft coordinates for the rect attribute.
        """
        self.prop_name = prop_name

        self.image = FrozenCover.prop_map[prop_name]

        self.rect = self.image.get_rect()

        self.rect.bottomleft = coordinates

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for key mapping management."""

from pygame import locals as constants
from pygame.key import name as name_key


DEFAULT_KEYS_MAP = {
    "player_left": constants.K_a,
    "player_right": constants.K_d,
    "player_up": constants.K_w,
    "player_down": constants.K_s,
    "jump": constants.K_k,
    "fire_item": constants.K_j,
    "interact": constants.K_e,
    "advance_deny": constants.K_SPACE,
    "accept_prompt": constants.K_RETURN,
}

VALID_KEYS_MAP = {
    name_key(eval("constants." + item)): eval("constants." + item)
    for item in dir(constants)
    if item.startswith("K_")
    if not "ALT" in item
    if not "SHIFT" in item
    if not "CTRL" in item
    if not "LOCK" in item
    if item != "K_ESCAPE"
}



from os import listdir, remove, mkdir
from os.path import join
from shutil import copyfile

### local imports

from ...logconfig import APP_LOGGER

from ...config import GAME_REFS, SAVES_DIR, STATE_FILE_NAME

from ...common.jsonhandler import load_json


logger = APP_LOGGER.getChild(__name__)


def load_game():
    """Return level name from the current save slot used."""
    logger.info("Loading a new save.")

    save_state_data = load_json(join(SAVES_DIR, GAME_REFS.dirname, STATE_FILE_NAME))

    last_level_name = save_state_data["last_level"]

    return last_level_name



"""Facility for managing music playback."""

from os import listdir
from os.path import join

### third-party import
from pygame.mixer import music

### local import
from .config import MUSIC_DIR, USER_SETTINGS


# TODO eliminate the "mood" feature. Always choose the
# music by name

# TODO export music manager methods individually, as done
# with the TASK_MANAGER


class MusicManager(object):
    """Manages music playback."""

    mood_to_music = {
        "casual": join(MUSIC_DIR, "caketown.ogg"),
        "tense": join(MUSIC_DIR, "title_screen_theme.ogg"),
    }

    def __init__(self, mood="tense"):
        """Assign variables and perform setups.

        mood (defaults to 'casual')
            A mood from DJ.moods.
        """
        if mood not in self.mood_to_music:
            raise ValueError("mood not listed in cls.mood_to_music")
        self.mood = mood

        self.play_mood()
        self.set_volume(USER_SETTINGS["music_volume"])

    def update(self):
        """Update object state.

        Do so by verifying if any music is being played
        and start playing if not."""
        busy = music.get_busy()
        if busy:
            pass
        else:
            self.play_mood()

    def cue(self, mood):
        """If needed, change music to match mood.

        mood
            A mood from DJ.moods."""
        if mood not in self.mood_to_music:
            raise ValueError("mood not listed in cls.mood_to_music")

        if self.mood == mood:
            pass
        else:
            self.mood = mood

            milliseconds = 1000
            music.fadeout(milliseconds)

    def play_mood(self):
        """Play the music for current mood."""
        music.load(self.mood_to_music[self.mood])

        loops = -1  # -1 means loop indefinitely
        music.play(loops)

    def get_volume(self):
        """Return music volume."""
        return music.get_volume()

    def set_volume(self, volume):
        """Set volume of music.

        volume
            Float from 0.0 to 1.0.
        """
        music.set_volume(volume)


MUSIC_MANAGER = MusicManager()



"""Facility for load screen management."""

from math import radians
from os.path import join
from itertools import cycle

### third-party imports

from pygame.time import wait, get_ticks as get_millisecs
from pygame.image import load as load_image
from pygame.display import update

### local imports

from ..config import SCREEN, SCREEN_RECT, MILLISECS_PER_FRAME, IMAGES_DIR

from ..palette import BLACK, WHITE
from ..logconfig import APP_LOGGER

from .text.main import render_text


logger = APP_LOGGER.getChild(__name__)


### class definition


class LoadScreenManager:
    """Common behaviour/objects for elegant screen loading."""

    def __init__(self):
        """Instantiate text objects, store millisec count."""
        ### create text objects to assist the class

        ## please wait text

        topright = SCREEN_RECT.move(-50, 35).topright

        self.wait_text = render_text(
            "Please wait...",
            font_size=36,
            foreground_color=BLACK,
            return_obj=True,
            coordinates_name="topright",
            coordinates_value=topright,
        )

        ## resources text

        center = SCREEN_RECT.move(0, 40).center

        self.resources_text = render_text(
            "Loading resources...",
            font_size=36,
            foreground_color=BLACK,
            return_obj=True,
            coordinates_name="center",
            coordinates_value=center,
        )

        self.millisecs_count = 0

        ### load surface objects for simple loading animation

        mushroom_filenames = (
            "mushroom_topleft.png",
            "mushroom_topright.png",
            "mushroom_bottomright.png",
            "mushroom_bottomleft.png",
        )

        surfs = [
            load_image(join(IMAGES_DIR, filename)).convert_alpha()
            for filename in mushroom_filenames
        ]

        ### store a function to indefinitely return the next
        ### surface from the list
        self.get_next_surf = cycle(surfs).__next__

        ### create a rect from one of the mushroom surfaces
        ### and position at the center of the screen

        self.mushroom_rect = self.get_next_surf().get_rect()
        self.mushroom_rect.center = SCREEN_RECT.center

    def blit_loading_screen(self):
        """Blit simple message while loading."""
        logger.info("Displaying loading screen.")

        SCREEN.fill((245, 245, 245))

        self.wait_text.draw()

        update()

        # wait around just half a second (500 milliseconds)
        # so the user can read the "Please wait" message;
        # this is just an usability measure; if we didn't
        # wait a bit, the text would disappear almost
        # immediately, since loading time is quick, and the
        # user might be left confused wondering whether the
        # text they couldn't read was important;
        wait(500)

    def blit_resource_loading(self):
        """Blit resource loading message as time passes."""
        ### calculate delta of time

        current_millisecs = get_millisecs()
        delta_t = current_millisecs - self.millisecs_count

        ### if delta time equals or surpasses the average
        ### millisecs per frame, blit resource loading
        ### message and update millisecs_count attribute

        if delta_t >= MILLISECS_PER_FRAME:

            SCREEN.fill((245, 245, 245))

            self.resources_text.draw()

            SCREEN.blit(self.get_next_surf(), self.mushroom_rect)

            update()

            self.millisecs_count = current_millisecs


load_screen_manager = LoadScreenManager()
blit_loading_screen = load_screen_manager.blit_loading_screen
blit_resource_loading = load_screen_manager.blit_resource_loading



"""Facility for specialized label for value monitoring."""

### local imports

from .main import Label

from ....palette import BLACK


class AutoLabel(Label):
    """A label with custom autoupdating ability."""

    def __init__(
        self,
        monitor_routine,
        prefix="",
        suffix="",
        text="",
        *,
        font_size=22,
        font_style="default",
        foreground_color=BLACK,
        background_color=(*BLACK, 0),
        coordinates_name="topleft",
        coordinates_value=(0, 0)
    ):
        """Perform setups and assign data for reuse.

        Extends appcommon.text.label.main.Label.__init__

        monitor_routine (callable)
            to be executed every loop to check for changes
            in its return value.
        prefix, suffix (string)
            text to be prepended/appended to the label text
            respectively.
        text (string)
            provides text to update surface.
        font_size (integer)
            indicates desired font size.
        font_style (string)
            Either 'default' for default font or 'mono' for
            a monospace font.
        foreground_color
        background_color
            A tuple or list of r, g, b values which are
            integers ranging from 0 to 255. The background
            color can be None, though, in which case the
            background is rendered transparent. For the
            background color, an optional fourth value can
            be passed, which is an integer in the same range
            as the others, representing the image opacity
            (0 for full transparency and 255 for full
            opacity).
        coordinates_name
            String representing attribute name of rect
            wherein to store the position information from
            the coordinates value parameter.
        coordinates_value
            Position information in the form of a tuple or
            list containing 2 integers representing
            positions in the x and y axes, respectively.
        """
        ### store variables for reuse

        self.monitor_routine = monitor_routine

        self.prefix = prefix
        self.suffix = suffix

        ### prepend prefix and append suffix to text
        text = self.prefix + text + self.suffix

        ### initialize superclass
        super().__init__(
            text,
            font_size=font_size,
            font_style=font_style,
            foreground_color=foreground_color,
            background_color=background_color,
            coordinates_name=coordinates_name,
            coordinates_value=coordinates_value,
        )

        ### override update behaviour from Label superclass
        self.update = self.monitor_content

        ### execute the update method, so the contents
        ### are update with the prefix and suffix if needed
        self.update()

    def monitor_content(self):
        """Pass input from monitor routine to Label.set.

        Uses appcommon.text.label.main.Label.set, which only
        updates the contents of the label if they have
        changed.
        """
        text = self.prefix + self.monitor_routine() + self.suffix

        self.set(text)



"""Facility for a options widgets."""

from functools import partialmethod

### third-party imports

from pygame.mouse import get_pos as get_mouse_pos, get_pressed as get_mouse_pressed
from pygame.draw import rect as draw_rect, line as draw_line

### local imports

from ...config import SCREEN

from ...common.behaviour import empty_function

from ...appcommon.style import give_depth_finish
from ...appcommon.text.main import render_text
from ...appcommon.surf import render_rect

from ...palette import BLACK, WHITE


def get_button(text, topleft, command):
    """Return a button-like object that can be invoked."""
    button = render_text(
        text, 32, return_obj=True, coordinates_value=topleft, foreground_color=WHITE
    )

    button.invoke = command
    button.image = give_depth_finish(button.image)

    # XXX rethink the design in the code below

    # pygame don't handle lambdas well, so using this
    def blit_button():
        """Simple function to blit button."""
        SCREEN.blit(button.image, button.rect)

    button.draw = blit_button
    button.update = empty_function

    return button


class VolumeScale(object):
    """A custom scale widgets from 0.0 to 1.0."""

    def __init__(self, mapping, field, topleft, update_command, feedback_command=None):
        """Assign variables and perform setups.

        mapping
            A dict whose specific field must be updated.
        field
            A string representing a field in a dict which
            must be update.
        topleft
            Topleft attribute for object rect.
        update_command
            Callable to be executed when updating the value.
        feedback_command
            A command to be executed after updating the value.
            Used to provide feedback. Defaults to None
        """
        self.mapping = mapping
        self.field = field
        self.update_command = update_command
        if feedback_command:
            self.feedback_command = feedback_command
        else:
            self.feedback_command = empty_function

        surf = render_rect(540, 50, BLACK)
        self.image = give_depth_finish(surf)
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

        draw_line(self.image, WHITE, (20, 25), (520, 25), 3)
        self.handle = render_rect(16, 16, WHITE, return_obj=True)
        self.place_handle()
        self.mouse_pressed = False

    def place_handle(self):
        """Place handle according to current value."""
        value = self.mapping[self.field]
        int_value = int(value * 100)
        offset_value = int_value * 5

        x = self.rect.x + 20 + offset_value
        y = self.rect.centery
        self.handle.rect.center = (x, y)

    def update(self):
        """Update object state."""
        self.check_handle()

    def draw(self):
        """Draw surfaces."""
        SCREEN.blit(self.image, self.rect)
        SCREEN.blit(self.handle.image, self.handle.rect)

    def check_handle(self):
        """Check if handle is being changed."""
        ### Assigning useful variables
        x, y = get_mouse_pos()
        widget_collided = self.rect.collidepoint(x, y)

        ### Checking scenarios with varying mouse states
        ### Remember: self.mouse_pressed is the state
        ### from previous frame
        if self.mouse_pressed and not widget_collided:
            self.update_value()

            # if mouse cursor leaves widget, even if it
            # is pressed, it must be considered released
            self.mouse_pressed = False
            return

        elif not self.mouse_pressed and not widget_collided:
            return

        # XXX the middle and right variables should be
        # assigned to '_' to indicate they won't be used.

        ### Get current mouse state and assign other variables
        left, middle, right = get_mouse_pressed()

        ### Branching different scenarios and their outcomes
        ### according to mouse state before and now.
        ### nomenclature: scenario -> outcome
        if self.mouse_pressed:
            self.mouse_pressed = left
            if self.mouse_pressed:
                ## pressed before and now -> position handle
                self.handle_to_x(x)
            else:
                ## pressed before and not now -> update value
                self.update_value()

        else:
            self.mouse_pressed = left
            if self.mouse_pressed:
                ## not before, pressed now -> position handle
                self.handle_to_x(x)

        # XXX
        # maybe the no/no scenario could have
        # the handle animated

    def handle_to_x(self, mouse_x):
        """Position handle according to mouse x coordinate.

        mouse_x
            Integer. Mouse x coordinate relative to screen
            and constrained to it.
        """
        left_edge = self.rect.x + 20
        right_edge = self.rect.x + self.rect.w - 20
        if left_edge <= mouse_x <= right_edge:
            self.handle.rect.centerx = mouse_x
        elif mouse_x > right_edge:
            self.handle.rect.centerx = self.rect.x + self.rect.w - 20
        elif mouse_x < left_edge:
            self.handle.rect.centerx = self.rect.x + 20

    def update_value(self):
        """Update mapping[field] using handle position."""
        offset = self.handle.rect.centerx - (self.rect.x + 20)
        value = offset / 500
        self.mapping[self.field] = value
        self.update_command()
        self.feedback_command()

    def increment_amount(self, amount):
        """Increment value according to amount.

        amount
            Any number that can be added to the volume value.
            The result will be clamped to the allowed
            interval [0.0, 1.0] (math notation).
        """
        value = self.mapping[self.field]
        new_value = value + amount
        clamped_value = max(0.0, min(1.0, new_value))
        self.mapping[self.field] = clamped_value
        self.place_handle()
        self.update_value()

    increase = partialmethod(increment_amount, 0.05)
    decrease = partialmethod(increment_amount, -0.05)



"""Facility for level additional groups and related setups.

Those groups are usually used for decorative purposes.
"""

from random import seed, choice, sample

### local imports

from ..appcommon.autoblit import BlitterSet

from ..cloud.clouds import Cloud
from ..mountain.mountains import Mountain


def set_additional_groups(level):
    """Set additional groups in game group maps.

    Such groups contain decorative objects, animals and
    plants which aren't part of the gameplay, but
    contribute to the aesthetics of the gaem.

    level
        Instance of gamelevel.main.Level.
    """
    seed()

    ### set mountains

    level.group_names.insert(0, "parallax_mountains")
    level.group_map["parallax_mountains"] = BlitterSet()
    level.onscreen_map["parallax_mountains"] = BlitterSet()

    mountain = Mountain()
    level.group_map["parallax_mountains"].add(mountain)
    level.objs_to_scroll.add(mountain)

    ### Set clouds

    level.group_names.insert(1, "parallax_clouds")
    level.group_map["parallax_clouds"] = BlitterSet()
    level.onscreen_map["parallax_clouds"] = BlitterSet()

    # XXX
    # The group insertion above could be a level's method,
    # since it's used repeatedly. Maybe not.

    # values chosen arbitrarily based on better
    # aesthetical composition
    cloud_bottomlefts = [[x, y] for x in (-160, 240, 640, 1040) for y in (-64, 192)]

    x_offsets = range(-20, 21, 5)
    y_offsets = range(-60, 61, 20)
    for item in cloud_bottomlefts:
        item[0] += choice(x_offsets)
        item[1] += choice(y_offsets)

    cloud_anim_names = ["idle{}".format(str(n).rjust(2, "0")) for n in range(1, 21)]

    random_cloud_anims = sample(cloud_anim_names, 8)

    cloud_data_pairs = zip(cloud_bottomlefts, random_cloud_anims)

    instances = [
        Cloud("cloud", bottomleft, cloud_anim_name)
        for bottomleft, cloud_anim_name in cloud_data_pairs
    ]

    level.group_map["parallax_clouds"].update(instances)
    level.objs_to_scroll.update(instances)

    #
    #    ### set trees
    #    level.group_names.insert(2, "parallax_trees")
    #    level.group_map["parallax_trees"]    = BlitterSet()
    #    level.onscreen_map["parallax_trees"] = BlitterSet()
    #
    #    ### set flora
    #    level.group_names.insert(5, "flora")
    #    level.group_map["animals"]    = BlitterSet()
    #    level.onscreen_map["animals"] = BlitterSet()
    #
    #    ### set fauna
    #    level.group_names.insert(6, "fauna")
    #    level.group_map["fauna"]    = BlitterSet()
    #    level.onscreen_map["fauna"] = BlitterSet()
    #
    #    ### set foreground
    #    level.group_names.append("parallax_fg")
    #    level.group_map["parallax_fg"]    = BlitterSet()
    #    level.onscreen_map["parallax_fg"] = BlitterSet()
    #
    ### get objects touching screen
    level.get_onscreen_data()



"""Facility for healthbar and health management."""

from pygame import Rect

from pygame.draw import rect as draw_rect

from ...config import SCREEN
from ...common.behaviour import empty_function
from ...common.math import unscroll_coordinates
from ...appcommon.surf import render_rect
from ...appcommon.autoblit import BasicObject
from ...palette import BLACK, ARBITRARY_COLORKEY


class HealthBar(BasicObject):
    """A healthbar for the thorn shooter."""

    def __init__(self, thorn_shooter):
        """Initialize superclass and perform setups."""
        self.thorn_shooter = thorn_shooter
        self.fill_color = (255, 90, 40)

        self.image = render_rect(100, 10, self.fill_color)
        self.rect = self.image.get_rect()
        self.rect.midbottom = unscroll_coordinates(self.thorn_shooter.rect.midtop)
        self.rect.y += -10

        self.image.set_colorkey(ARBITRARY_COLORKEY)
        self.stencil = render_rect(*self.image.get_size(), ARBITRARY_COLORKEY)

        self.set_operational_structures()

    def set_operational_structures(self):
        """Define more data and perform additional setups."""
        x, y = self.rect.topleft
        outline_x = x - 2
        outline_y = y - 2

        width, height = self.image.get_size()
        outline_width = width + 2
        outline_height = height + 2

        self.outline_rect = Rect(
            (outline_x, outline_y), (outline_width, outline_height)
        )
        self.max_health = 100
        self.min_health = 0

        self.update_width()

        if self.thorn_shooter.health:
            self.draw = self.draw_alive

        else:
            self.draw = empty_function

    def update_width(self, amount=0):
        """Increase or decrease health by amount.

        It also updates size according to current health.

        amount (defaults to 0)
            An integer expressing the amount to add or
            subtract (if it is negative) from health.
        """
        if not self.thorn_shooter.health:
            return

        self.thorn_shooter.health += amount
        self.thorn_shooter.health = max(
            min(self.thorn_shooter.health, self.max_health), self.min_health
        )

        self.image.fill(self.fill_color)
        self.image.blit(self.stencil, (self.thorn_shooter.health, 0))

        if not self.thorn_shooter.health:
            self.thorn_shooter.die()
            self.draw = empty_function

    def draw_alive(self):
        """Draw healthbar (used only when alive."""
        super().draw()
        draw_rect(SCREEN, BLACK, self.outline_rect, 2)

    def revive(self):
        """Resets healthbar state to alive."""
        self.draw = self.draw_alive
        self.update_width()



"""Facility for the main menu class and related tools."""

from os import listdir
from os.path import isdir, join
from collections import deque
from functools import partialmethod


### third-party imports

from pygame import (
    QUIT,
    KEYDOWN,
    K_RETURN,
    K_w,
    K_UP,
    K_s,
    K_DOWN,
    MOUSEBUTTONDOWN,
    MOUSEMOTION,
)

from pygame.display import update
from pygame.event import get as get_events
from pygame.draw import rect as draw_rect


### local imports

from ..logconfig import APP_LOGGER
from ..config import SCREEN, SCREEN_RECT, GAME_REFS, SAVES_DIR, STATE_FILE_NAME

from ..common.jsonhandler import load_json

from ..appcommon.task import update_task_manager
from ..appcommon.surf import render_rect, render_image
from ..appcommon.style import give_depth_finish
from ..appcommon.autoblit import BlitterSet, BasicObject
from ..appcommon.exception import (
    QuitGameException,
    LevelSwitchException,
    ManagerSwitchException,
)
from ..appcommon.text.main import render_text
from ..appcommon.behaviour.load import load_game

from ..font import get_font
from ..sound import SOUNDS_MAP
from ..music import MUSIC_MANAGER
from ..palette import CERULEAN, BLACK, WHITE, SKY_COLOR

from .newgame import NewGameWidget
from .loadgame import LoadGameWidget
from .options.main import OptionsWidget


logger = APP_LOGGER.getChild(__name__)


class MainMenu:
    """A group-like class for a main menu sprites."""

    def __init__(self):
        """Set this instance."""
        ### handle music
        MUSIC_MANAGER.cue("tense")

        ### gather conditions

        save_folder_exists = isdir(SAVES_DIR)
        it_is_not_empty = listdir(SAVES_DIR)

        ### evaluate then

        if save_folder_exists and it_is_not_empty:

            self.widget_names = [
                "Continue",
                "New game",
                "Load game",
                "Options",
                "Credits",
                "Exit game",
            ]

            self.load_game_widget = LoadGameWidget(self)

        else:

            self.widget_names = ["New game", "Options", "Credits", "Exit game"]

        self.new_game_widget = NewGameWidget(self)
        self.options_widget = OptionsWidget(self)

        # XXX
        # code the credits screen
        # as well as the respective main menu widget method
        self.credits_screen = None

        self.background = render_rect(*SCREEN_RECT.size, color=SKY_COLOR)

        self.game_logo = render_image(
            "danittr.png", return_obj=True, coordinates_value=(660, 100)
        )

        self.build_items()

    def build_items(self):
        """Create menu item widgets."""
        ### Create menu widget group and its buttons
        self.menu_widgets = deque()

        x, y = 250, 120
        for name in self.widget_names:
            ### Chose command
            if name == "Continue":
                command = self.load_last_save
            elif name == "New game":
                command = self.create_new_game
            elif name == "Load game":
                command = self.load_game
            elif name == "Options":
                command = self.present_options
            elif name == "Credits":
                command = self.play_credits
            elif name == "Exit game":
                command = self.exit_game

            ### Instantiate and set up

            menu_widget = render_text(
                name,
                font_size=38,
                foreground_color=WHITE,
                background_color=BLACK,
                return_obj=True,
                coordinates_value=(x, y),
            )

            menu_widget.image = give_depth_finish(menu_widget.image)

            menu_widget.invoke = command
            self.menu_widgets.append(menu_widget)
            y += 70

        ### Create group to help manage blitting
        self.widgets_group = BlitterSet(self.menu_widgets)

        self.selected_item = self.menu_widgets[0]
        self.highlight_rect = self.selected_item.rect

    def control(self):
        """Update menu state based on event_queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            ### Keyboard
            elif event.type == KEYDOWN:

                if event.key in (K_w, K_UP):
                    self.select_previous()

                elif event.key in (K_s, K_DOWN):
                    self.select_next()

                elif event.key == K_RETURN:
                    self.selected_item.invoke()

            ### Mouse
            elif event.type == MOUSEMOTION:
                self.mouse_motion_routine(event.pos)

            elif event.type == MOUSEBUTTONDOWN:

                if event.button == 1:
                    self.mouse_click_routine(event.pos)
                self.mouse_motion_routine(event.pos)

    def update(self):
        """Update the task manager."""
        update_task_manager()

    def draw(self):
        """Draw menu and draw its items."""
        SCREEN.blit(self.background, (0, 0))
        SCREEN.blit(self.game_logo.image, self.game_logo.rect)

        self.widgets_group.draw()
        draw_rect(SCREEN, CERULEAN, self.highlight_rect, 4)

        update()

    def rotate_selection(self, amount=0):
        """Assign selected item after rotating by amount."""
        self.menu_widgets.rotate(amount)
        self.selected_item = self.menu_widgets[0]
        self.highlight_rect = self.selected_item.rect
        SOUNDS_MAP["gui_step"].play()

    select_previous = partialmethod(rotate_selection, 1)
    select_next = partialmethod(rotate_selection, -1)

    def mouse_motion_routine(self, mouse_position):
        """Check if mouse touches any widget and select it.

        mouse_position
            Tuple with x and y integers representing
            mouse position coordinates relative to
            screen topleft corner and constrained to
            the screen.
        """
        ### Search widgets
        for item in self.menu_widgets:
            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break
        else:
            touched_item = None

        ### If already selected, we are done
        if touched_item == self.selected_item:
            return

        ### Select widget + admin tasks
        if touched_item:
            for i in range(len(self.menu_widgets)):
                if touched_item == self.menu_widgets[0]:
                    self.rotate_selection()
                    break
                self.menu_widgets.rotate(1)

    def mouse_click_routine(self, mouse_position):
        """Invoke widget if mouse click over it.

        mouse_position
            Tuple with x and y integers representing
            mouse position coordinates relative to
            screen topleft corner and constrained to
            the screen.
        """
        ### Search widget
        for item in self.menu_widgets:
            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break
        else:
            touched_item = None

        ### Select item if needed, then invoke it
        if touched_item:
            if touched_item != self.selected_item:
                self.mouse_motion_routine(mouse_position)
            touched_item.invoke()

    def load_last_save(self):
        """Load most recent save file."""
        ### Define variables
        last_date = ""
        respective_dir = ""

        ### Iterate over dirs to determine most recent one
        for directory in listdir(SAVES_DIR):

            state_path = join(SAVES_DIR, directory, STATE_FILE_NAME)

            data = load_json(state_path)

            str_datetime = data["last_played_datetime"]

            if not last_date:
                last_date = str_datetime
                respective_dir = directory

            else:
                if max(last_date, str_datetime) != last_date:
                    last_date = str_datetime
                    respective_dir = directory

        ### Play sound + start game

        SOUNDS_MAP["game_start"].play()

        GAME_REFS.dirname = respective_dir

        level_name = load_game()

        raise LevelSwitchException(level_name)

    def create_new_game(self):
        """Load the new game setup."""
        raise ManagerSwitchException(self.new_game_widget)

    def load_game(self):
        """Load a list of save game from which to choose."""
        raise ManagerSwitchException(self.load_game_widget)

    def present_options(self):
        """Present the options setup."""
        raise ManagerSwitchException(self.options_widget)

    def play_credits(self):
        """Present the game credits."""
        # XXX
        # To be coded

    def exit_game(self):
        """Exit game."""
        logger.debug("Exiting game from main menu.")
        raise QuitGameException



"""Facility for tablet menu classes and tools."""

# XXX
# This class is currently just a placeholder class created
# with escape menu old code. It should be properly coded soon.

from collections import deque

### third-party imports

from pygame import Surface, QUIT, KEYDOWN, K_RETURN, K_ESCAPE, K_w, K_s, K_m

from pygame.display import update
from pygame.event import get as get_events

### local imports

from ..config import SCREEN

from ..appcommon.autoblit import BlitterSet, BasicObject

from ..font import get_font
from ..palette import BLACK, WHITE


class TabletMenu(object):
    """Presents a tablet menu."""

    def __init__(self, level):
        """Initialize instance."""
        self.font = get_font(32)
        self.level = level

        self.widget_names = [
            "Resume game",
            "Save game",
            "Quit to main menu",
            "Quit to desktop",
        ]

        self.set_background()
        self.build_items()
        self.set_arrow_widget()

    def set_background(self):
        """Set a background for the menu."""
        self.background = Surface((400, 420)).convert()
        self.background.fill(BLACK)

    def build_items(self):
        """Create menu item widgets."""
        self.widgets_group = BlitterSet()
        self.menu_widgets = deque()
        x_coordinate, y_coordinate = 620, 320
        for name in self.widget_names:
            self.menu_widgets.append(
                TabletWidget(name, self, self.level, (x_coordinate, y_coordinate))
            )
            y_coordinate += 50

        self.selected_item = self.menu_widgets[0]
        self.widgets_group.update(self.menu_widgets)

    def set_arrow_widget(self):
        """Create a highlight arrow for selected items."""
        self.arrow_widget = BasicObject()
        self.arrow_widget.image = self.font.render(">", True, WHITE, BLACK)
        self.arrow_widget.rect = self.arrow_widget.image.get_rect()
        x, y = self.selected_item.rect.topleft
        self.arrow_widget.rect.topleft = (x - 20, y)
        self.widgets_group.add(self.arrow_widget)

    def control(self):
        """Handle events in level event queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            elif event.type == KEYDOWN:

                if event.key == K_w:
                    self.select_previous()

                elif event.key == K_s:
                    self.select_next()

                elif event.key == K_RETURN:
                    self.invoke_item_action()

                elif event.key in (K_ESCAPE, K_m):
                    raise ManagerSwitchException(self.level)

    def update(self):
        """Update menu state."""
        x, y = self.selected_item.rect.topleft
        self.arrow_widget.rect.topleft = (x - 20, y)

    def draw(self):
        """Draw objects."""
        SCREEN.blit(self.background, (580, 260))
        self.widgets_group.draw()

        update()

    def select_previous(self):
        """Select previous menu item."""
        self.menu_widgets.rotate()
        self.selected_item = self.menu_widgets[0]

    def select_next(self):
        """Select next menu item."""
        self.menu_widgets.rotate(-1)
        self.selected_item = self.menu_widgets[0]

    def invoke_item_action(self):
        """Invoke the action on selected menu item."""
        self.selected_item.action()


class TabletWidget(object):
    """Escape menu item widget."""

    def __init__(self, name, escape_menu, level, topleft):
        """Initialize superclass and set instance."""
        self.font = get_font(32)

        self.name = name
        self.escape_menu = escape_menu
        self.level = level

        if name == "Resume game":
            self.action = self.resume_game
        if name == "Save game":
            self.action = self.save_game
        if name == "Quit to main menu":
            self.action = self.quit_to_main_menu
        elif name == "Quit to desktop":
            self.action = self.exit_game

        self.image = self.font.render(name, True, WHITE, BLACK)
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

    def resume_game(self):
        """Resume game."""
        raise ManagerSwitchException(self.level)

    def save_game(self):
        """Save game and level states and give feedback."""

    def quit_to_main_menu(self):
        """Load main menu."""

    def exit_game(self):
        """Exit game."""



"""Facility for action controlling during gameplay."""

from pygame.event import pump


class ControlOverrider:
    """An object to control and/or record gameplay actions.

    It should be used to control game objects actions
    for various purposes.

    E.g.: to create in-game cutscenes; or to record
    action sequences for playtesting or to simulate
    AI/animate objects.

    By now it can only play predefined scripts, but
    the goal is to make it read custom formated data
    either provided by the user or recorded from
    object actions.

    The mechanism used is fairly simple. The
    gamelevel.main.Level instance let's the
    overrider override it's event handling/key input
    handling methods, accepting actions performed by
    the overrider. The rest of the game objects behaves
    either naturally by their own update mechanisms or
    they also may have such mechanisms overridden by
    the overrider, depending on your need of control.

    Control swap between the level instance and overrider
    may be performed by the overrider itself or by other
    objects depending on conditions or timing. The class
    already works, but it is too 'green' yet and there's
    still other user cases to be implement. Thus, this class
    must be expected to change a lot while I incrementally
    develop/redesign it over and over, along with its
    dependencies.
    """

    def __init__(self):
        """Create object."""
        self.actions_map = {}

    def register_actions(self, key, actions):
        """Store actions in the action map."""
        self.actions_map[key] = actions

    def define_script(self, script):
        """Define a script (detailed steps) for execution.

        Such script must explain the actions to be performed
        from the moment the overrider takes over until it
        gives control back to the level.

        The control switching isn't necessarily present
        in the instruction set and may be performed by
        other objects, specially when events are 'chained'
        together (the execution of one is only triggered
        after the execution of a previous one), in which
        case providing a step by step 'guide' isn't needed,
        just the first one is enough (since it will
        automatically call further steps).

        script
            Either a string hinting predefined actions or
            all actions to be performed. For now only the
            string case is being treated.
        """
        if script == "stand_still":
            self.keys = ("player", "stop")

    def prepare_script(self):
        """Store action in the execute script attribute."""
        key1, key2 = self.keys
        self.execute_script = self.actions_map[key1][key2]

    def perform_actions(self):
        """Perform scripted actions.

        Such actions consist of the execution of
        instance methods of the objects the overrider is
        controlling according to a script (script as in
        performance/plays, not as in programming).
        """
        ### Administrative task to keep event queue "healthy"
        pump()
        self.execute_script()


OVERRIDER = ControlOverrider()

register_actions = OVERRIDER.register_actions
define_script = OVERRIDER.define_script
prepare_script = OVERRIDER.prepare_script
perform_actions = OVERRIDER.perform_actions


# XXX
# When the class expands enough,
# remember to clear its "cache" whenever
# switching levels.



"""Mushroom item class facility."""

from ..logconfig import APP_LOGGER
from ..appcommon.anim.player.main import AnimationPlayer


class ToxicPollenCloud:
    """A circular-shaped toxic cloud of pollen.

    It deals a little damage in touching actors or player.
    It exists as an attribute of a Mushroom instance until
    said instance hits the ground as a projectile. When this
    happens, this pollen cloud enters the game world
    via the level method 'add_obj_to_group'.
    """

    level = None

    def __init__(self, prop_name, coordinates, level):
        """Initialize object."""
        self.prop_name = prop_name
        self.anim_player = AnimationPlayer(
            self, prop_name, coordinates_name="topleft", coordinates_value=coordinates
        )

        self.set_operation_structure()

    def set_operation_structure(self):
        """Set other important operational structures.

        Those include atributes, methods and other objects."""
        self.damage = 1
        self.facing_right = True

        self.anim_player.switch_animation("appearing")
        self.anim_player.play()

    def prepare_to_enter(self, mushroom):
        """Make it so it's ready to enter game world."""
        self.rect.center = mushroom.rect.center
        self.facing_right = mushroom.facing_right

    def update(self):
        """Execute current update_method method."""
        self.manage_animations()
        self.anim_player.update()
        if self.anim_player.anim_name in ("idle", "idle_reverse"):
            self.damage_live_objects()

    def draw(self):
        """Draw animation."""
        self.anim_player.draw()

    def manage_animations(self):
        """Select suitable animation according to state."""
        anim_name = self.anim_player.anim_name

        if anim_name == "appearing":

            loops_no = self.anim_player.peek_next_loops_no()

            if loops_no >= 1:

                if self.facing_right:
                    self.anim_player.switch_animation("idle")
                else:
                    self.anim_player.switch_animation("idle_reverse")

        elif anim_name in ("idle", "idle_reverse"):
            loops_no = self.anim_player.peek_next_loops_no()
            if loops_no >= 3:
                self.anim_player.switch_animation("disappearing")

        elif anim_name == "disappearing":
            loops_no = self.anim_player.peek_after_next_loops_no()
            if loops_no >= 1:
                self.level.remove_obj_from_group(self, "front_props", True)

    def damage_live_objects(self):
        """Damage close objects."""
        colliding_sprites = []

        colliding_actors_onscreen = self.level.onscreen_map["actors"].collide(self)
        colliding_sprites.extend(colliding_actors_onscreen)

        player = self.level.player
        if self.rect.colliderect(player.rect):
            colliding_sprites.append(player)

        for sprite in colliding_sprites:

            try:
                sprite.suffer_damage(self.damage)

            except AttributeError:
                # XXX
                # filter this better for the abovementioned
                # method. Also add logging (maybe?!).
                pass

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for surface generation for general purposes."""

from itertools import cycle, chain
from collections.abc import Sequence

### third-party imports

from pygame import Surface, Rect
from pygame.transform import flip as flip_surface
from pygame.draw import line as draw_line

### local imports

from .surf import INVISIBLE_SURF, render_image
from .loadscreen import blit_resource_loading


def generate_surfaces(surf_data, surf_map={}):
    """Return surfaces list after series of operations.

    surf_data (dict or list)
        data used to generate/load/alter data recursively
        or not.
    surf_map (dict)
        contains surface lists. It is only meant to provide
        values and never be altered here.
    """
    ### process operation to
    ### retrieve/load/generate/alter surfaces
    name = surf_data["name"]

    ## retrieve from surf_map
    if name == "retrieve_from_map":
        key = surf_data["key"]
        output = surf_map[key]

    ## load from disk
    elif name == "load_from_disk":
        image_names = surf_data["image_names"]

        output = []

        for img_name in image_names:

            ### admin task: blit a loading resource message
            blit_resource_loading()

            ### load and append surf

            surf = render_image(img_name)
            output.append(surf)

    elif name == "load_spritesheet":
        img_name = surf_data["image_name"]

        ### admin task: blit a loading resource message
        blit_resource_loading()

        ### load spritesheet
        spritesheet = render_image(img_name)

        output = []

        try:
            sprite_size = surf_data["sprite_size"]

        ## use specified rects to retrieve each sprite
        except KeyError:
            rects_parameters = surf_data["rects_parameters"]
            for rect_params in rects_parameters:
                rect = Rect(rect_params)
                subsurf = spritesheet.subsurface(rect)
                output.append(subsurf)

        ## otherwise use sprite size and quantity data
        ## to retrieve each sprite by walking the
        ## spritesheet like a grid/table
        else:
            quantity = surf_data["how_many_sprites"]
            sprite_width, sprite_height = sprite_size
            spritesheet_width = spritesheet.get_width()
            sprites_per_row = spritesheet_width // sprite_width

            # create a cycle to count in which row
            # you are when iterating
            row_counter = cycle(range(sprites_per_row))

            row = next(row_counter)
            column = 0
            for _ in range(quantity):
                x = sprite_width * row
                y = sprite_height * column

                # create rect and use it to create
                # subsurf
                rect = Rect(x, y, *sprite_size)
                subsurf = spritesheet.subsurface(rect)

                # store subsurf
                output.append(subsurf)

                # get next row;
                # if it equals zero, increment column;
                row = next(row_counter)
                if row == 0:
                    column += 1

    ## use an invisible surf
    elif name == "invisible":
        output = INVISIBLE_SURF

    ## create an special surf similar to an "empty"
    ## object in blender software
    elif name == "create_empty":

        surf = Surface(surf_data["dimensions"]).convert_alpha()

        surf.fill((0, 0, 0, 0))

        rect = surf.get_rect()

        lines = [[rect.midtop, rect.midbottom], [rect.midleft, rect.midright]]

        for line in lines:
            draw_line(surf, (0, 0, 0), *line, 3)

        output = surf

    ## change surf list by slicing it
    elif name == "slice":
        surfaces = generate_surfaces(surf_data["surface_data"], surf_map)

        start = surf_data.get("start")
        stop = surf_data.get("stop")
        step = surf_data.get("step")

        output = surfaces[start:stop:step]

    ## change surf list by flipping its surfaces
    elif name == "flip":
        surfaces = generate_surfaces(surf_data["surface_data"], surf_map)

        flip_x = surf_data.get("flip_x", False)
        flip_y = surf_data.get("flip_y", False)

        output = [flip_surface(surf, flip_x, flip_y) for surf in surfaces]

    ## change surf list by deepcopying surfs
    elif name == "deepcopy":
        surfaces = generate_surfaces(surf_data["surface_data"], surf_map)

        output = [surf.copy() for surf in surfaces]

    ## concatenate lists of surfaces
    elif name == "concatenate":

        # retrieve list of surface sequences data
        surface_sequences = surf_data["surface_sequences"]

        # generate a list of surfaces for each surface data

        lists = [
            generate_surfaces(surface_data, surf_map)
            for surface_data in surface_sequences
        ]

        # finally concatenate all lists into a single one
        output = list(chain.from_iterable(lists))

    ### make sure return value is a sequence

    if isinstance(output, Sequence):
        return output

    else:
        return [output]



"""Facility for loading game on main menu."""

from os import listdir
from types import SimpleNamespace
from os.path import join
from functools import partialmethod

### third-party imports

from pygame import (
    QUIT,
    KEYDOWN,
    K_RETURN,
    K_ESCAPE,
    K_w,
    K_UP,
    K_s,
    K_DOWN,
    MOUSEBUTTONDOWN,
    MOUSEMOTION,
)

from pygame.display import update
from pygame.event import get as get_events
from pygame.image import load as load_image
from pygame.draw import rect as draw_rect
from pygame.mouse import get_pos as get_mouse_pos

### local imports

from ..config import SCREEN, SCREEN_RECT, GAME_REFS, SAVES_DIR, STATE_FILE_NAME

from ..common.math import get_reaching_multiple
from ..common.behaviour import empty_function
from ..common.jsonhandler import load_json
from ..common.wdeque.main import WalkingDeque

from ..appcommon.task import add_task, update_task_manager
from ..appcommon.surf import render_rect
from ..appcommon.style import give_depth_finish
from ..appcommon.autoblit import BlitterSet
from ..appcommon.exception import (
    QuitGameException,
    LevelSwitchException,
    ManagerSwitchException,
)
from ..appcommon.text.main import render_text
from ..appcommon.behaviour.load import load_game

from ..sound import SOUNDS_MAP
from ..palette import CERULEAN, BLACK, WHITE, SKY_COLOR


SLOT_DISTANCE = 110

# XXX
# Now, besides the art, the only thing the mainmenu subpackage
# lacks is mouse style and animation. This would be a fine
# touch.


class LoadGameWidget(object):
    """Widget to help loading new games."""

    def __init__(self, main_menu):
        """Set the instance.

        main_menu
            Reference of mainmenu.main.MainMenu instance.
        """
        self.main_menu = main_menu

        self.thumb = render_rect(640, 360, return_obj=True, coordinates_value=(630, 25))

        self.build_slot_list()

        self.background = render_rect(*SCREEN_RECT.size, color=SKY_COLOR)

        self.set_labels()

        self.error_cooldown_task = SimpleNamespace(finished=True)

    def set_labels(self):
        """Set a labels group to display labels."""
        self.labels = BlitterSet()

        font_size = 28

        text01 = "Please,"
        label01 = render_text(
            text01,
            font_size,
            foreground_color=WHITE,
            padding=0,
            return_obj=True,
            coordinates_value=(25, 25),
        )

        text02 = "select a save:"
        label02 = render_text(
            text02,
            font_size,
            foreground_color=WHITE,
            padding=0,
            return_obj=True,
            coordinates_value=(25, 60),
        )

        self.labels.update([label01, label02])

    def build_slot_list(self):
        """Build each individual slot with info."""
        ### Collect saves information
        saves = []
        for dirname in listdir(SAVES_DIR):
            path = join(SAVES_DIR, dirname, STATE_FILE_NAME)
            last_played = load_json(path)["last_played_datetime"]
            save = SimpleNamespace(dirname=dirname, last_played=last_played)
            saves.append(save)

        ### Sort by last played, generate visual slots
        # pygame don't handle lambdas well, so using this.
        def get_last_played(item):
            """Return last played attribute value."""
            return item.last_played

        saves.sort(key=get_last_played, reverse=True)

        x, y = 200, 25
        slots = []
        for save in saves:
            topleft = (x, y)
            slot = SaveSlot(save.dirname, save.last_played, topleft)
            slots.append(slot)
            y += SLOT_DISTANCE

        ### Create groups of such slots to help manage them
        self.slots_deque = WalkingDeque(slots)
        self.slots = BlitterSet(self.slots_deque)

        ### Define choosen one and rect to highlight it
        self.selected_slot = self.slots_deque[0]
        self.highlight_rect = self.selected_slot.rect
        self.thumb.image = self.selected_slot.thumb

    def control(self):
        """Update widget state based on event_queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            ### Keyboard

            elif event.type == KEYDOWN:

                if event.key in (K_w, K_UP):
                    self.select_previous()
                elif event.key in (K_s, K_DOWN):
                    self.select_next()
                elif event.key == K_RETURN:
                    self.load_game_from_slot()
                elif event.key == K_ESCAPE:
                    raise ManagerSwitchException(self.main_menu)

            ### Mouse
            elif event.type == MOUSEMOTION:
                self.mouse_motion_routine(event.pos)

            elif event.type == MOUSEBUTTONDOWN:

                if event.button == 1:
                    self.mouse_click_routine(event.pos)
                elif event.button == 4:
                    self.scroll_down()
                elif event.button == 5:
                    self.scroll_up()

    def update(self):
        """Update the task manager."""
        update_task_manager()

    def draw(self):
        """Draw widget objects."""
        SCREEN.blit(self.background, (0, 0))
        self.labels.draw()

        self.slots.draw()
        draw_rect(SCREEN, CERULEAN, self.highlight_rect, 4)

        SCREEN.blit(self.thumb.image, self.thumb.rect)
        update()

    def rotate_selection(self, amount=0, from_mouse=False):
        """Assign selected item after rotating by amount.

        amount
            Integer representing amount of deque rotation.
        from_mouse
            Boolean. If true, scroll check isn't performed,
            since it is expected that the user perform
            the scrolling itself with the mouse wheel.
        """
        self.slots_deque.walk(amount)
        self.selected_slot = self.slots_deque[0]
        self.highlight_rect = self.selected_slot.rect
        self.thumb.image = self.selected_slot.thumb

        if not from_mouse:
            self.check_scrolling()

        SOUNDS_MAP["gui_step"].play()

    select_previous = partialmethod(rotate_selection, -1)
    select_next = partialmethod(rotate_selection, 1)

    def mouse_motion_routine(self, mouse_position):
        """Check if mouse touches any widget and select it.

        mouse_position
            Tuple containing mouse coordinates relative to
            screen topleft and always constrained to it.
        """
        ### Search slot

        for item in self.slots_deque:

            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break

        else:
            touched_item = None

        ### If already selected, get out
        if touched_item == self.selected_slot:
            return

        ### Select slot + admin tasks

        if touched_item:

            for i in range(len(self.slots_deque)):

                if touched_item == self.slots_deque[0]:
                    self.rotate_selection(from_mouse=True)
                    break

                self.slots_deque.walk(1)

    def mouse_click_routine(self, mouse_position):
        """Invoke widget if mouse click over it.

        mouse_position
            Tuple containing mouse coordinates relative to
            screen topleft and always constrained to it.
        """
        ### Search slot

        for item in self.slots_deque:
            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break

        else:
            touched_item = None

        ### Select slot if not already + load slot

        if touched_item:

            if touched_item != self.selected_slot:
                self.mouse_motion_routine(mouse_position)

            self.load_game_from_slot()

    def scroll(self, amount):
        """Scroll slots by amount.

        amount
            Integer representing amount to be added to
            rect.y attribute, which is how scrolling is done.
        """
        ### Refuse to scroll up if first slot is on screen

        if amount > 0:

            first_slot = self.slots_deque.get_mem_value(0)

            if SCREEN_RECT.contains(first_slot.rect):
                self.play_error()
                return

        ### Refuse to scroll down if last slot is on screen

        elif amount < 0:

            last_slot = self.slots_deque.get_mem_value(-1)

            if SCREEN_RECT.contains(last_slot.rect):
                self.play_error()
                return

        ### Otherwise, scroll and check if mouse hovers any
        for slot in self.slots:
            slot.rect.y += amount

        self.mouse_motion_routine(get_mouse_pos())

    scroll_up = partialmethod(scroll, -SLOT_DISTANCE)
    scroll_down = partialmethod(scroll, SLOT_DISTANCE)

    def play_error(self):
        """Time and execute play error appropriately."""
        if self.error_cooldown_task.finished:
            SOUNDS_MAP["error"].play()
            self.error_cooldown_task = add_task(empty_function, 800)

    def check_scrolling(self):
        """Check if current slot needs to be scrolled.

        This is to be sure the selected slot is aways
        on screen.
        """
        ### Store reusable values
        slot_top = self.selected_slot.rect.top
        slot_bottom = self.selected_slot.rect.bottom
        screen_top = SCREEN_RECT.top
        screen_bottom = SCREEN_RECT.bottom

        ### How much scrolling if slot is bellow screen
        if slot_top < screen_top:
            distance = screen_top - slot_top
            to_scroll = get_reaching_multiple(SLOT_DISTANCE, distance)
            self.scroll(to_scroll)

        ### How much scrolling if slot is above screen
        elif slot_bottom > screen_bottom:
            distance = slot_bottom - screen_bottom
            to_scroll = get_reaching_multiple(SLOT_DISTANCE, distance)
            self.scroll(-to_scroll)

    def load_game_from_slot(self):
        """Load the game from slot dirname attribute."""
        SOUNDS_MAP["game_start"].play()

        GAME_REFS.dirname = self.selected_slot.dirname

        level_name = load_game()

        raise LevelSwitchException(level_name)


class SaveSlot:
    """Container representing a save slot."""

    font_size = 28

    # XXX
    # 2018-01-12 10:29:12
    # Include last level/section on the slots blited info
    # And also a screenshot that must be saved on dirname
    # whenever saving the game within the interactive levels.

    def __init__(self, dirname, last_played, topleft):
        """Save slot."""
        self.dirname = dirname
        self.save_name = dirname.replace("save_", "", 1)
        self.last_played = last_played

        surf = render_rect(420, 80)
        self.image = give_depth_finish(surf)
        self.rect = self.image.get_rect()
        self.rect.topleft = topleft

        self.blit_details()
        self.load_thumb()

    def blit_details(self):
        """Blit the final details on the image surface."""
        save_name_text = "Name:  {}".format(self.save_name)
        save_name_surf = render_text(
            save_name_text,
            font_size=self.font_size,
            foreground_color=WHITE,
            background_color=BLACK,
            padding=0,
        )

        last_played_text = "Last played:  {}".format(self.last_played)
        last_played_surf = render_text(
            last_played_text,
            font_size=self.font_size,
            foreground_color=WHITE,
            background_color=BLACK,
            padding=0,
        )

        self.image.blit(save_name_surf, (5, 1))
        self.image.blit(last_played_surf, (5, 35))

    def load_thumb(self):
        """Load and store thumb surface.

        Such surface is referenced by other objects to
        blit it in the appropriate location.
        """
        path = join(SAVES_DIR, self.dirname, "thumb.png")

        self.thumb = load_image(path).convert()



"""Facility for sign middle prop class."""

### local imports

from ..common.behaviour import empty_function
from ..appcommon.autoblit import BasicObject
from ..appcommon.signfactory import sign_factory


class Sign(BasicObject):
    """A class to represent a sign or signpost.

    The surfaces are generated in
    appcommon/signfactory.py
    """

    #     _______        _______|\         _______|\
    #    |       |      |         \       |         \
    #    | Hello |  or  | Hello    \  or  | Hello    \
    #    |_______|      |          /      |          /
    #                   |_______  /       |_______  /
    #                           |/           |  | |/
    #                                        |  |
    #                                        |__|

    level = None

    def __init__(
        self,
        prop_name,
        coordinates,
        text,
        fg_color,
        bg_color,
        signpost_height=0,
        direction=None,
    ):
        """Initialize superclass and variables.

        prop_name
            A string representing the prop name.
        coordinates
            List or tuple of x and y coordinates representing
            the desired bottomleft coordinates for the
            rect attribute.
        text
            A string to be used on the sign surface.
        fg_color
            The foreground color of the sign. A tuple
            or list with three integers representing
            r, g, b values from 0 to 255.
        bg_color
            The background color of the sign. A tuple
            or list with three integers representing
            r, g, b values from 0 to 255.
        signpost_height
            Integer representing the height of the
            post which supports the sign. If 0 there will
            be no such post.
        direction
            A string of the set {'n', 's', 'e', 'w'} or
            None. If none, the sign won't be represented
            by an arrow pointing in a specific direction,
            but by a rectangle.
        """
        self.prop_name = prop_name

        self.image = sign_factory(text, fg_color, bg_color, signpost_height, direction)

        self.rect = self.image.get_rect()
        self.rect.bottomleft = coordinates

        if signpost_height:
            self.rect.y += 4

        self.update = empty_function

    def scroll(self, x, y):
        """Scroll by x and y in respective axis.
        x
            Integer indicating scrolling in x axis.
        y
            Integer indicating scrolling in y axis.
        """
        self.rect.move_ip(x, y)



"""Facility for options configuration.

Here we provide an object which manages the screen.
It's purpose is to present the player with means to
change settings in the USER_SETTING mapping and have
them applied.
"""

from collections import deque
from functools import partialmethod

### third-party imports

from pygame import (
    QUIT,
    KEYDOWN,
    K_ESCAPE,
    K_RETURN,
    K_w,
    K_UP,
    K_s,
    K_DOWN,
    K_a,
    K_LEFT,
    K_d,
    K_RIGHT,
    MOUSEBUTTONDOWN,
    MOUSEMOTION,
)

from pygame.display import update
from pygame.event import get as get_events
from pygame.draw import rect as draw_rect

### local imports

from ...config import SCREEN, SCREEN_RECT, DEFAULT_SETTINGS, USER_SETTINGS

from ...appcommon.task import update_task_manager
from ...appcommon.surf import render_rect
from ...appcommon.autoblit import BlitterSet
from ...appcommon.exception import QuitGameException, ManagerSwitchException
from ...appcommon.text.main import render_text

from ...appcommon.behaviour.setting import apply_user_settings

from .widgets import VolumeScale, get_button
from .controls import ControlsWidget

from ...sound import SOUNDS_MAP
from ...palette import CERULEAN, BLACK, WHITE, SKY_COLOR


class OptionsWidget(object):
    """Widget for setting up a new game and its data.

    The user is prompted for a name for a new save."""

    def __init__(self, main_menu):
        """Set variables and perform setups.

        main_menu
            The instance of the mainmenu.main.MainMenu class.
        """
        self.main_menu = main_menu

        self.set_labels()
        self.background = render_rect(*SCREEN_RECT.size, color=SKY_COLOR)
        self.control_widget = ControlsWidget(self)
        self.build_widgets()

    def set_labels(self):
        """Create a sprite to serve as a label."""
        self.labels = BlitterSet()

        text01 = "Music Volume"
        label01 = render_text(
            text01,
            32,
            foreground_color=WHITE,
            background_color=BLACK,
            return_obj=True,
            coordinates_value=(100, 130),
        )

        text02 = "SFX Volume"
        label02 = render_text(
            text02,
            32,
            foreground_color=WHITE,
            background_color=BLACK,
            return_obj=True,
            coordinates_value=(100, 250),
        )

        self.labels.update([label01, label02])

    def build_widgets(self):
        """Create option widgets."""
        ### Create container and define variable
        self.widgets_deque = deque()

        update_command = apply_user_settings

        ### Music volume
        self.widgets_deque.append(
            VolumeScale(USER_SETTINGS, "music_volume", (100, 180), update_command)
        )

        jump_sound = SOUNDS_MAP["player_jump01"]

        ### Sound volume
        self.widgets_deque.append(
            VolumeScale(
                USER_SETTINGS,
                "sound_volume",
                (100, 300),
                update_command,
                jump_sound.play,
            )
        )

        ### Restore defaults
        self.widgets_deque.append(
            get_button("Restore volume defaults", (100, 400), self.restore_defaults)
        )

        ### Configure controls
        self.widgets_deque.append(
            get_button("Configure controls", (100, 520), self.switch_to_control)
        )

        ### Create group to help manage blitting
        self.widgets = BlitterSet(self.widgets_deque)

        ### Select first widget and rect to be highlighted
        self.current_widget = self.widgets_deque[0]
        self.highlight_rect = self.current_widget.rect

    def control(self):
        """Update widget state based on event_queue."""
        for event in get_events():

            if event.type == QUIT:
                raise QuitGameException

            ### Keyboard

            elif event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    raise ManagerSwitchException(self.main_menu)

                elif event.key == K_RETURN:
                    self.invoke_button()

                elif event.key in (K_w, K_UP):
                    self.go_up()

                elif event.key in (K_s, K_DOWN):
                    self.go_down()

                elif event.key in (K_a, K_LEFT):

                    try:
                        self.current_widget.decrease()
                    except AttributeError:
                        pass

                elif event.key in (K_d, K_RIGHT):

                    try:
                        self.current_widget.increase()
                    except AttributeError:
                        pass

            ### Mouse

            elif event.type == MOUSEMOTION:
                self.mouse_motion_routine(event.pos)

            elif event.type == MOUSEBUTTONDOWN:

                if event.button == 1:
                    self.mouse_click_routine(event.pos)

    def update(self):
        """Update widgets states."""
        self.widgets.update_objs()
        update_task_manager()

    def draw(self):
        """Draw widgets and update screen."""
        SCREEN.blit(self.background, (0, 0))

        self.labels.draw()
        self.widgets.draw_objs()

        draw_rect(SCREEN, CERULEAN, self.highlight_rect, 3)

        update()

    def switch_widget(self, steps=0):
        """Switch widgets by steps.

        steps
            integers representing how many the widgets deque
            should rotate.
        """
        self.widgets_deque.rotate(steps)
        self.current_widget = self.widgets_deque[0]

        self.highlight_rect = self.current_widget.rect
        SOUNDS_MAP["gui_step"].play()

    go_up = partialmethod(switch_widget, 1)
    go_down = partialmethod(switch_widget, -1)

    def mouse_motion_routine(self, mouse_position):
        """Check if mouse touches any widget and select it.

        mouse_position
            Tuple with x and y integers representing
            mouse position relative to the topleft corner
            of the screen, constrained to the screen.
        """
        ### Search widget
        for item in self.widgets_deque:
            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break
        else:
            touched_item = None

        ### If already selected, we are done
        if touched_item == self.current_widget:
            return

        ### Select widget
        if touched_item:
            for i in range(len(self.widgets_deque)):
                if touched_item == self.widgets_deque[0]:
                    self.switch_widget()
                    break
                self.widgets_deque.rotate(1)

    def mouse_click_routine(self, mouse_position):
        """Invoke widget if mouse click over it.

        mouse_position
            Tuple with x and y integers representing
            mouse position relative to the topleft corner
            of the screen, constrained to the screen.
        """
        ### Search widget
        for item in self.widgets_deque:
            if item.rect.collidepoint(mouse_position):
                touched_item = item
                break
        else:
            touched_item = None

        ### Select it if needed then invoke it
        if touched_item:
            if touched_item != self.current_widget:
                self.mouse_motion_routine(mouse_position)
            self.invoke_button()

    def restore_defaults(self):
        """Restore volume defaults."""
        USER_SETTINGS["music_volume"] = DEFAULT_SETTINGS["music_volume"]

        USER_SETTINGS["sound_volume"] = DEFAULT_SETTINGS["sound_volume"]

        apply_user_settings()

        for widget in self.widgets_deque:

            try:
                widget.place_handle()
            except AttributeError:
                pass

    def switch_to_control(self):
        """Switch update manager to control widget."""
        raise ManagerSwitchException(self.control_widget)

    def invoke_button(self):
        """Invoke current widget, if it can be invoked."""
        try:
            self.current_widget.invoke()
        except AttributeError:
            pass
